import sys
import json
import asyncio
import logging

import zigpy_znp.types as t
from zigpy_znp.api import ZNP
from zigpy_znp.config import CONFIG_SCHEMA
from zigpy_znp.exceptions import SecurityError, CommandNotRecognized
from zigpy_znp.types.nvids import NWK_NVID_TABLES, ExNvIds, OsalNvIds
from zigpy_znp.tools.common import ClosableFileType, setup_parser

LOGGER = logging.getLogger(__name__)


async def nvram_read(znp: ZNP):
    data = {}
    data["LEGACY"] = {}

    # Legacy items need to be handled first, since they are named
    for nwk_nvid in OsalNvIds:
        if nwk_nvid == OsalNvIds.INVALID_INDEX:
            continue

        # Tables span ranges of items. Naming them properly is useful.
        if nwk_nvid in NWK_NVID_TABLES:
            start = nwk_nvid
            end = NWK_NVID_TABLES[nwk_nvid]

            for offset in range(0, end - start):
                key = f"{nwk_nvid.name}+{offset}"

                try:
                    value = await znp.nvram.osal_read(
                        nwk_nvid + offset, item_type=t.Bytes
                    )
                except SecurityError:
                    LOGGER.warning("Read disallowed for %s", key)
                    continue
                except KeyError:
                    break

                LOGGER.info("%s = %r", key, value.hex())
                data["LEGACY"][key] = value.hex()
        else:
            try:
                value = await znp.nvram.osal_read(nwk_nvid, item_type=t.Bytes)
            except KeyError:
                continue
            except SecurityError:
                LOGGER.warning("Read disallowed for %s", nwk_nvid)
                continue

            LOGGER.info("%s = %r", nwk_nvid, value.hex())
            data["LEGACY"][nwk_nvid.name] = value.hex()

    for nvid in ExNvIds:
        # Skip the LEGACY items, we did them above
        if nvid == ExNvIds.LEGACY:
            continue

        for sub_id in range(2 ** 16):
            try:
                value = await znp.nvram.read(
                    item_id=nvid, sub_id=sub_id, item_type=t.Bytes
                )
            except CommandNotRecognized:
                # CC2531 only supports the legacy NVRAM interface, even on Z-Stack 3
                return data
            except KeyError:
                # Once a read fails, no later reads will succeed
                break

            LOGGER.info("%s[0x%04X] = %r", nvid.name, sub_id, value.hex())
            data.setdefault(nvid.name, {})[f"0x{sub_id:04X}"] = value.hex()

    return data


async def main(argv):
    parser = setup_parser("Backup a radio's NVRAM")
    parser.add_argument(
        "--output", "-o", type=ClosableFileType("w"), help="Output file", default="-"
    )

    args = parser.parse_args(argv)

    with args.output as f:
        znp = ZNP(CONFIG_SCHEMA({"device": {"path": args.serial}}))
        await znp.connect()

        obj = await nvram_read(znp)
        znp.close()

        f.write(json.dumps(obj, indent=4) + "\n")


if __name__ == "__main__":
    asyncio.run(main(sys.argv[1:]))  # pragma: no cover



"""This interface allows the tester to interact with the target at system level such
as reset, read/write memory, read/write extended address, etc.
"""

import zigpy_znp.types as t


class BootloaderBuildType(t.enum_uint8):
    NON_BOOTLOADER_BUILD = 0
    BUILT_AS_BIN = 1
    BUILT_AS_HEX = 2


class ADCChannel(t.enum_uint8):
    """The ADC channel."""

    AIN0 = 0x00
    AIN1 = 0x01
    AIN2 = 0x02
    AIN3 = 0x03
    AIN4 = 0x04
    AIN5 = 0x05
    AIN6 = 0x06
    AIN7 = 0x07
    Temperature = 0x0E
    Voltage = 0x0F


class ADCResolution(t.enum_uint8):
    """Resolution of the ADC channel."""

    bits_8 = 0x00
    bits_10 = 0x01
    bits_12 = 0x02
    bits_14 = 0x03


class GPIOPinMode(t.enum_flag_uint8):
    """Pin state. Any pin with an unspecified state bit is pulled up."""

    Tristate0 = 0b0000_0001
    Tristate1 = 0b0000_0010
    Tristate2 = 0b0000_0100
    Tristate3 = 0b0000_1000

    PullDown0 = 0b0001_0000
    PullDown1 = 0b0010_0000
    PullDown2 = 0b0100_0000
    PullDown3 = 0b1000_0000


class GPIOPinDirection(t.enum_flag_uint8):
    """Pin direction. Any pin with an unspecified direction bit is an input pin."""

    Output0 = 0b0000_0001
    Output1 = 0b0000_0010
    Output2 = 0b0000_0100
    Output3 = 0b0000_1000


class GPIOOperation(t.enum_uint8):
    """Specifies the type of operation to perform on the GPIO pins."""

    SetDirection = 0x00
    SetTristateMode = 0x01
    Set = 0x02
    Clear = 0x03
    Toggle = 0x04
    Read = 0x05
    HiD = 0x12  # ???


class StackTuneOperation(t.enum_uint8):
    """The tuning operation to be executed."""

    # XXX: [Value] should correspond to the valid values specified by the
    # ZMacTransmitPower_t enumeration (0xFD - 0x16)
    PowerLevel = 0x00

    # Set RxOnWhenIdle off/on if the value of Value is 0/1;
    # otherwise return the 0x01 current setting of RxOnWhenIdle.
    SetRxOnWhenIdle = 0x01


class SYS(t.CommandsBase, subsystem=t.Subsystem.SYS):
    # reset the target device
    ResetReq = t.CommandDef(
        t.CommandType.AREQ,
        0x00,
        req_schema=(
            t.Param(
                "Type",
                t.ResetType,
                (
                    "This command will reset the device by using a hardware reset "
                    "(i.e. watchdog reset) if 'Type' is zero. Otherwise a soft "
                    "reset (i.e. a jump to the reset vector) is done. This "
                    "is especially useful in the CC2531, for instance, so that the "
                    "USB host does not have to contend with the USB H/W resetting "
                    "(and thus causing the USB host to reenumerate the device "
                    "which can cause an open virtual serial port to hang.)"
                ),
            ),
        ),
    )

    # issue PING requests to verify if a device is active and check the capability of
    # the device
    Ping = t.CommandDef(
        t.CommandType.SREQ,
        0x01,
        req_schema=(),
        rsp_schema=(
            t.Param(
                "Capabilities",
                t.MTCapabilities,
                "Represents the intefaces this device can handle",
            ),
        ),
    )

    # request for the device's version string
    Version = t.CommandDef(
        t.CommandType.SREQ,
        0x02,
        req_schema=(),
        rsp_schema=(
            t.Param("TransportRev", t.uint8_t, "Transport protocol revision"),
            t.Param("ProductId", t.uint8_t, "Product ID"),
            t.Param("MajorRel", t.uint8_t, "Software major release number"),
            t.Param("MinorRel", t.uint8_t, "Software minor release number"),
            t.Param("MaintRel", t.uint8_t, "Software maintenance release number"),
            # Optional stuff
            t.Param(
                "CodeRevision",
                t.uint32_t,
                "User-supplied code revision number",
                optional=True,
            ),
            t.Param(
                "BootloaderBuildType",
                BootloaderBuildType,
                "Bootloader build type",
                optional=True,
            ),
            t.Param(
                "BootloaderRevision",
                t.uint32_t,
                "Bootloader revision. 0 - not provided, 0xFFFFFFFF - not supported",
                optional=True,
            ),
        ),
    )

    # set the extended address of the device
    SetExtAddr = t.CommandDef(
        t.CommandType.SREQ,
        0x03,
        req_schema=(t.Param("ExtAddr", t.EUI64, "The device's extended address"),),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # get the extended address of the device
    GetExtAddr = t.CommandDef(
        t.CommandType.SREQ,
        0x04,
        req_schema=t.STATUS_SCHEMA,
        rsp_schema=(t.Param("ExtAddr", t.EUI64, "The device's extended address"),),
    )

    # read a single memory location in the target RAM. The command accepts an address
    # value and returns the memory value present in the target RAM at that address
    RamRead = t.CommandDef(
        t.CommandType.SREQ,
        0x05,
        req_schema=(
            t.Param("Address", t.uint16_t, "Address of the memory to read"),
            t.Param("Len", t.uint8_t, "The number of bytes to read"),
        ),
        rsp_schema=(
            t.Param("Status", t.Status, "Status is either Success (0) or Failure (1)"),
            t.Param("Value", t.ShortBytes, "The value read from memory address"),
        ),
    )

    # write to a particular location in the target RAM. The command accepts an
    # address location and a memory value. The memory value is written to the address
    # location in the target RAM
    RamWrite = t.CommandDef(
        t.CommandType.SREQ,
        0x06,
        req_schema=(
            t.Param("Address", t.uint16_t, "Address of the memory to read"),
            t.Param("Value", t.ShortBytes, "The value read from memory address"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # create and initialize an item in non-volatile memory. The NV item will be created
    # if it does not already exist. The data for the new NV item will be left
    # uninitialized if the InitLen parameter is zero. When InitLen is non-zero, the
    # data for the NV item will be initialized (starting at offset of zero) with the
    # values from InitData. Note that it is not necessary to initialize the entire NV
    # item (InitLen < ItemLen). It is also possible to create an NV item that is larger
    # than the maximum length InitData - use the SYS_OSAL_NV_WRITE command to finish
    # the initialization
    OSALNVItemInit = t.CommandDef(
        t.CommandType.SREQ,
        0x07,
        req_schema=(
            t.Param("Id", t.uint16_t, "The Id of the NV Item"),
            t.Param("ItemLen", t.uint16_t, "Number of bytes in the NV item"),
            t.Param("Value", t.ShortBytes, "The value of the NV item"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # read a single memory item in the target non-volatile memory. The command accepts
    # an attribute Id value and returns the memory value present in the target for the
    # specified attribute Id
    OSALNVRead = t.CommandDef(
        t.CommandType.SREQ,
        0x08,
        req_schema=(
            t.Param("Id", t.uint16_t, "The Id of the NV item"),
            t.Param(
                "Offset",
                t.uint8_t,
                "Number of bytes offset from the beginning of the NV value",
            ),
        ),
        rsp_schema=(
            t.Param("Status", t.Status, "Status is either Success (0) or Failure (1)"),
            t.Param("Value", t.ShortBytes, "The value of the NV item"),
        ),
    )

    # write to a particular item in non-volatile memory. The command accepts an
    # attribute Id and an attribute value. The attribute value is written to the
    # location specified for the attribute Id in the target
    OSALNVWrite = t.CommandDef(
        t.CommandType.SREQ,
        0x09,
        req_schema=(
            t.Param("Id", t.uint16_t, "The Id of the NV item"),
            t.Param(
                "Offset",
                t.uint8_t,
                "Number of bytes offset from the beginning of the NV value",
            ),
            t.Param("Value", t.ShortBytes, "The value of the NV item"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # delete an item from the non-volatile memory. The ItemLen parameter must match
    # the length of the NV item or the command will fail
    OSALNVDelete = t.CommandDef(
        t.CommandType.SREQ,
        0x12,
        req_schema=(
            t.Param("Id", t.uint16_t, "The Id of the NV item"),
            t.Param("ItemLen", t.uint16_t, "Number of bytes in the NV item"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # get the length of an item in non-volatile memory. A returned length of zero
    # indicates that the NV item does not exist
    OSALNVLength = t.CommandDef(
        t.CommandType.SREQ,
        0x13,
        req_schema=(t.Param("Id", t.uint16_t, "The Id of the NV item"),),
        rsp_schema=(t.Param("ItemLen", t.uint16_t, "Number of bytes in the NV item"),),
    )

    SetJammerParameters = t.CommandDef(
        t.CommandType.SREQ,
        0x15,
        req_schema=(
            t.Param(
                "ContinuousEvents",
                t.uint16_t,
                "Number of continuous events needed to detect Jamming",
            ),
            t.Param("HighNoiseLevel", t.uint8_t, "Noise Level needed to be a Jam"),
            t.Param(
                "DetectPeriodTime",
                t.uint32_t,
                "The time between each noise level reading",
            ),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # start a timer event. The event will expired after the indicated amount of time
    # and a notification will be sent back to the tester
    OSALStartTimer = t.CommandDef(
        t.CommandType.SREQ,
        0x0A,
        req_schema=(
            t.Param("Id", t.uint8_t, "The Id of the timer event (0-3)"),
            t.Param("Timeout", t.uint16_t, "Timer timeout in millliseconds"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # stop a timer event
    OSALStopTimer = t.CommandDef(
        t.CommandType.SREQ,
        0x0B,
        req_schema=(t.Param("Id", t.uint8_t, "The Id of the timer event (0-3)"),),
        rsp_schema=t.STATUS_SCHEMA,
    )

    #  get a random 16-bit number
    Random = t.CommandDef(
        t.CommandType.SREQ,
        0x0C,
        req_schema=(),
        rsp_schema=(t.Param("Value", t.uint16_t, "The random value"),),
    )

    # read a value from the ADC based on specified channel and resolution
    ADCRead = t.CommandDef(
        t.CommandType.SREQ,
        0x0D,
        req_schema=(
            t.Param("Channel", ADCChannel, "The channel of ADC to read"),
            t.Param(
                "Resolution",
                ADCResolution,
                "Resolution of the reading: 8/10/12/14 bits",
            ),
        ),
        rsp_schema=(t.Param("Value", t.uint16_t, "Value of ADC channel"),),
    )

    # control the 4 GPIO pins on the CC2530-ZNP build
    GPIO = t.CommandDef(
        t.CommandType.SREQ,
        0x0E,
        req_schema=(
            t.Param(
                "Operation",
                GPIOOperation,
                "Specifies type of operation on GPIO pins",
            ),
            t.Param("Value", t.uint8_t, "GPIO value for specified operation"),
        ),
        rsp_schema=(t.Param("Value", t.uint8_t, "GPIO value"),),
    )

    # tune intricate or arcane settings at runtime
    StackTune = t.CommandDef(
        t.CommandType.SREQ,
        0x0F,
        req_schema=(
            t.Param(
                "Operation",
                StackTuneOperation,
                "Specifies type of operation on GPIO pins",
            ),
            t.Param("Value", t.uint8_t, "Tuning value"),
        ),
        rsp_schema=(
            (t.Param("Value", t.uint8_t, "Applicable status of the tuning operation"),)
        ),
    )

    # set the target system date and time. The time can be specified in
    # "seconds since 00:00:00 on January 1, 2000"
    # or in parsed date/time components
    SetTime = t.CommandDef(
        t.CommandType.SREQ,
        0x10,
        req_schema=(
            t.Param(
                "UTCTime",
                t.uint32_t,
                "Number of seconds since 00:00:00 on Jan 2000",
            ),
            t.Param("Hour", t.uint8_t, "Hour of the day (0 -- 23)"),
            t.Param("Minute", t.uint8_t, "Minute of the hour (0 -- 59)"),
            t.Param("Second", t.uint8_t, "Seconds of the minute (0 -- 59)"),
            t.Param("Month", t.uint8_t, "Month of the year (1 -- 12)"),
            t.Param("Day", t.uint8_t, "Day of the month (1 -- 31)"),
            t.Param("Year", t.uint16_t, "Year (2000 -- )"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # get the target system date and time. The time is returned in
    # "seconds since 00:00:00 on January 1, 2000" and parsed date/time components
    GetTime = t.CommandDef(
        t.CommandType.SREQ,
        0x11,
        req_schema=(),
        rsp_schema=(
            t.Param(
                "UTCTime",
                t.uint32_t,
                "Number of seconds since 00:00:00 on Jan 2000",
            ),
            t.Param("Hour", t.uint8_t, "Hour of the day (0 -- 23)"),
            t.Param("Minute", t.uint8_t, "Minute of the hour (0 -- 59)"),
            t.Param("Second", t.uint8_t, "Seconds of the minute (0 -- 59)"),
            t.Param("Month", t.uint8_t, "Month of the year (1 -- 12)"),
            t.Param("Day", t.uint8_t, "Day of the month (1 -- 31)"),
            t.Param("Year", t.uint16_t, "Year (2000 -- )"),
        ),
    )

    # set the target system radio transmit power. The returned TX power is the actual
    # setting applied to the radio - nearest characterized value for the specific
    # radio
    SetTxPower = t.CommandDef(
        t.CommandType.SREQ,
        0x14,
        req_schema=(t.Param("TXPower", t.int8s, "Requested TX power setting, in dBm"),),
        # XXX: Z-Stack 3.30+ returns SUCCESS or INVALID_PARAMETER.
        #      Z-Stack 1.2 and 3.0 return the cloest TX power setting.
        rsp_schema=(
            t.Param("StatusOrPower", t.int8s, "Status code or applied power setting"),
        ),
    )

    # initialize the statistics table in NV memory
    ZDiagsInitStats = t.CommandDef(
        t.CommandType.SREQ, 0x17, req_schema=(), rsp_schema=t.STATUS_SCHEMA
    )

    # clear the statistics table. To clear data in NV (including the Boot
    # Counter) the clearNV flag shall be set to TRUE
    ZDiagsClearStats = t.CommandDef(
        t.CommandType.SREQ,
        0x18,
        req_schema=(
            t.Param("ClearNV", t.Bool, "True -- clear statistics in NV memory"),
        ),
        rsp_schema=(t.Param("SycClock", t.uint32_t, "Milliseconds since last reset"),),
    )

    # read a specific system (attribute) ID statistics and/or metrics value
    ZDiagsGetStats = t.CommandDef(
        t.CommandType.SREQ,
        0x19,
        req_schema=(
            # as defined in ZDiags.h
            t.Param("AttributeId", t.uint16_t, "System diagnostics attribute ID"),
        ),
        rsp_schema=(t.Param("Value", t.uint32_t, "Value of the requested attribute"),),
    )

    # restore the statistics table from NV into the RAM table
    ZDiagsRestoreStatsNV = t.CommandDef(
        t.CommandType.SREQ, 0x1A, req_schema=(), rsp_schema=t.STATUS_SCHEMA
    )

    # save the statistics table from RAM to NV
    ZDiagsSaveStatsToNV = t.CommandDef(
        t.CommandType.SREQ,
        0x1B,
        req_schema=(),
        rsp_schema=(t.Param("SycClock", t.uint32_t, "Milliseconds since last reset"),),
    )

    # Same as OSALNVRead but with a 16-bit offset
    OSALNVReadExt = t.CommandDef(
        t.CommandType.SREQ,
        0x1C,
        req_schema=(
            t.Param("Id", t.uint16_t, "The Id of the NV item"),
            t.Param(
                "Offset",
                t.uint16_t,
                "Number of bytes offset from the beginning of the NV value",
            ),
        ),
        rsp_schema=(
            t.Param("Status", t.Status, "Status is either Success (0) or Failure (1)"),
            t.Param("Value", t.ShortBytes, "The value of the NV item"),
        ),
    )

    # Same as OSALNVWrite but with a 16-bit offset
    OSALNVWriteExt = t.CommandDef(
        t.CommandType.SREQ,
        0x1D,
        req_schema=(
            t.Param("Id", t.uint16_t, "The Id of the NV item"),
            t.Param(
                "Offset",
                t.uint16_t,  # XXX: don't trust the documentation! This *not* 8 bits.
                "Number of bytes offset from the beginning of the NV value",
            ),
            t.Param("Value", t.LongBytes, "The value of the NV item"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # attempt to create an item in non-volatile memory
    NVCreate = t.CommandDef(
        t.CommandType.SREQ,
        0x30,
        req_schema=(
            t.Param("SysId", t.uint8_t, "System ID of the NV item"),
            t.Param("ItemId", t.uint16_t, "Item ID of the NV item"),
            t.Param("SubId", t.uint16_t, "Sub ID of the NV item"),
            t.Param("Length", t.uint32_t, "Number of bytes in the NV item"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # attempt to delete an item in non-volatile memory
    NVDelete = t.CommandDef(
        t.CommandType.SREQ,
        0x31,
        req_schema=(
            t.Param("SysId", t.uint8_t, "System ID of the NV item"),
            t.Param("ItemId", t.uint16_t, "Item ID of the NV item"),
            t.Param("SubId", t.uint16_t, "Sub ID of the NV item"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # get the length of an item in non-volatile memory
    NVLength = t.CommandDef(
        t.CommandType.SREQ,
        0x32,
        req_schema=(
            t.Param("SysId", t.uint8_t, "System ID of the NV item"),
            t.Param("ItemId", t.uint16_t, "Item ID of the NV item"),
            t.Param("SubId", t.uint16_t, "Sub ID of the NV item"),
        ),
        rsp_schema=(t.Param("Length", t.uint32_t, "Length of NV item"),),
    )

    # read an item in non-volatile memory
    NVRead = t.CommandDef(
        t.CommandType.SREQ,
        0x33,
        req_schema=(
            t.Param("SysId", t.uint8_t, "System ID of the NV item"),
            t.Param("ItemId", t.uint16_t, "Item ID of the NV item"),
            t.Param("SubId", t.uint16_t, "Sub ID of the NV item"),
            t.Param(
                "Offset",
                t.uint16_t,
                "Number of bytes offset from the beginning of the NV value",
            ),
            t.Param("Length", t.uint8_t, "Length of data to read"),
        ),
        rsp_schema=(
            t.Param("Status", t.Status, "Status is either Success (0) or Failure (1)"),
            t.Param("Value", t.ShortBytes, "Value of the NV item read"),
        ),
    )

    # write an item in non-volatile memory
    NVWrite = t.CommandDef(
        t.CommandType.SREQ,
        0x34,
        req_schema=(
            t.Param("SysId", t.uint8_t, "System ID of the NV item"),
            t.Param("ItemId", t.uint16_t, "Item ID of the NV item"),
            t.Param("SubId", t.uint16_t, "Sub ID of the NV item"),
            t.Param(
                "Offset",
                t.uint16_t,
                "Number of bytes offset from the beginning of the NV value",
            ),
            # XXX: the spec has length as a 16-bit integer but then shows it as
            # an 8-bit integer in the table below, which matches the code
            t.Param("Value", t.ShortBytes, "Value of the NV item to write"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # update an item in non-volatile memory
    NVUpdate = t.CommandDef(
        t.CommandType.SREQ,
        0x35,
        req_schema=(
            t.Param("SysId", t.uint8_t, "System ID of the NV item"),
            t.Param("ItemId", t.uint16_t, "Item ID of the NV item"),
            t.Param("SubId", t.uint16_t, "Sub ID of the NV item"),
            t.Param("Value", t.ShortBytes, "Value of the NV item to update"),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # compact the active page in non-volatile memory
    NVCompact = t.CommandDef(
        t.CommandType.SREQ,
        0x36,
        req_schema=(
            t.Param(
                "Threshold",
                t.uint16_t,
                "Compaction occurs when NV bytes are less than this value",
            ),
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # MT SYS Callbacks
    # This command is sent by the device to indicate the reset
    ResetInd = t.CommandDef(
        t.CommandType.AREQ,
        0x80,
        rsp_schema=(
            t.Param("Reason", t.ResetReason, "Reason for the reset"),
            t.Param("TransportRev", t.uint8_t, "Transport protocol revision"),
            t.Param("ProductId", t.uint8_t, "Product ID"),
            t.Param("MajorRel", t.uint8_t, "Software major release number"),
            t.Param("MinorRel", t.uint8_t, "Software minor release number"),
            t.Param("MaintRel", t.uint8_t, "Software maintenance release number"),
        ),
    )

    # This command is sent by the device to indicate a specific time has been expired
    OSALTimerExpired = t.CommandDef(
        t.CommandType.AREQ,
        0x81,
        rsp_schema=(t.Param("Id", t.uint8_t, "The Id of the timer event (0-3)"),),
    )

    JammerInd = t.CommandDef(
        t.CommandType.AREQ,
        0x82,
        rsp_schema=(
            t.Param(
                "JammerInd",
                t.Bool,
                "TRUE if jammer detected, FALSE if changed to undetected",
            ),
        ),
    )



import zigpy_znp.types as t

Z2M_PAN_ID = 0xA162
Z2M_EXT_PAN_ID = t.EUI64.convert("DD:DD:DD:DD:DD:DD:DD:DD")
Z2M_NETWORK_KEY = t.KeyData([1, 3, 5, 7, 9, 11, 13, 15, 0, 2, 4, 6, 8, 10, 12, 13])

DEFAULT_TC_LINK_KEY = t.TCLinkKey(
    ExtAddr=t.EUI64.convert("FF:FF:FF:FF:FF:FF:FF:FF"),  # global
    Key=t.KeyData(b"ZigBeeAlliance09"),
    TxFrameCounter=0,
    RxFrameCounter=0,
)
ZSTACK_CONFIGURE_SUCCESS = t.uint8_t(0x55)

EMPTY_ADDR_MGR_ENTRY_ZSTACK1 = t.AddrMgrEntry(
    type=t.AddrMgrUserType.Default,
    nwkAddr=0xFFFF,
    extAddr=t.EUI64.convert("FF:FF:FF:FF:FF:FF:FF:FF"),
)

EMPTY_ADDR_MGR_ENTRY_ZSTACK3 = t.AddrMgrEntry(
    type=t.AddrMgrUserType(0xFF),
    nwkAddr=0xFFFF,
    extAddr=t.EUI64.convert("FF:FF:FF:FF:FF:FF:FF:FF"),
)

EMPTY_KEY = t.NwkKeyDesc(
    KeySeqNum=0x00,
    Key=t.KeyData([0x00] * 16),
)

# Used only when creating a temporary network during formation
STARTUP_CHANNELS = t.Channels.from_channel_list([15, 20, 25])



import logging
import itertools

import zigpy_znp.types as t
import zigpy_znp.commands as c
from zigpy_znp.types import nvids
from zigpy_znp.exceptions import SecurityError, InvalidCommandResponse

LOGGER = logging.getLogger(__name__)


# Some NVIDs don't really exist and Z-Stack doesn't behave consistently when operations
# are performed on them.
PROXIED_NVIDS = {nvids.OsalNvIds.POLL_RATE_OLD16}


class NVRAMHelper:
    def __init__(self, znp):
        self.znp = znp
        self.align_structs = None

    async def determine_alignment(self) -> None:
        """
        Automatically determine struct memory alignment. Must be called before any
        structs are read/written.
        """

        # TODO: Figure out a way to use `c.UTIL.AssocFindDevice.Req(Index=0)`!
        #       This is the only (known) MT command to just send a struct's in-memory
        #       representation over serial.

        # NWKKEY is almost always present, regardless of adapter state.
        # It is an 8-bit sequence number, a 16 byte key, and a 32-bit frame counter.
        # This is at least 1 + 16 + 4 = 21 bytes, or 24 if you have to 32-bit align.
        value = await self.osal_read(nvids.OsalNvIds.NWKKEY, item_type=t.Bytes)

        if len(value) == 24:
            self.align_structs = True
        elif len(value) == 21:
            self.align_structs = False
        else:
            raise ValueError(f"Unexpected value for NWKKEY: {value!r}")

        LOGGER.debug("Detected struct alignment: %s", self.align_structs)

    def serialize(self, value) -> bytes:
        """
        Serialize an object, automatically computing struct padding based on the target
        platform.
        """

        if hasattr(value, "serialize"):
            if isinstance(value, (t.CStruct, t.BaseListType)):
                assert self.align_structs is not None
                value = value.serialize(align=self.align_structs)
            else:
                value = value.serialize()
        elif not isinstance(value, (bytes, bytearray)):
            raise TypeError(
                f"Only bytes or serializable types can be written to NVRAM."
                f" Got {value!r} (type {type(value)})"
            )

        if not value:
            raise ValueError("NVRAM value cannot be empty")

        return value

    def deserialize(self, data: bytes, item_type, *, allow_trailing=False):
        """
        Serialize raw bytes, automatically computing struct padding based on the target
        platform.
        """

        if issubclass(item_type, (t.CStruct, t.BaseListType)):
            assert self.align_structs is not None
            value, remaining = item_type.deserialize(data, align=self.align_structs)
        else:
            value, remaining = item_type.deserialize(data)

        if remaining and not allow_trailing:
            raise ValueError(f"Data left after deserialization: {remaining!r}")

        return value

    async def osal_delete(self, nv_id: t.uint16_t) -> bool:
        """
        Deletes an item from NVRAM. Returns whether or not the item existed.
        """

        length = (await self.znp.request(c.SYS.OSALNVLength.Req(Id=nv_id))).ItemLen

        if length == 0:
            return False

        delete_rsp = await self.znp.request(
            c.SYS.OSALNVDelete.Req(Id=nv_id, ItemLen=length)
        )

        return delete_rsp.Status == t.Status.SUCCESS

    async def osal_write(self, nv_id: t.uint16_t, value, *, create: bool = False):
        """
        Writes a complete value to NVRAM, optionally resizing and creating the item if
        necessary.

        Serializes all serializable values and passes bytes directly.
        """

        value = self.serialize(value)
        length = (await self.znp.request(c.SYS.OSALNVLength.Req(Id=nv_id))).ItemLen

        # Recreate the item if the length is not correct
        if length != len(value) and nv_id not in PROXIED_NVIDS:
            if not create:
                if length == 0:
                    raise KeyError(f"NV item does not exist: {nv_id!r}")
                else:
                    raise ValueError(
                        f"Stored length and actual length differ:"
                        f" {length} != {len(value)}"
                    )

            if length != 0:
                await self.znp.request(
                    c.SYS.OSALNVDelete.Req(Id=nv_id, ItemLen=length),
                    RspStatus=t.Status.SUCCESS,
                )

            await self.znp.request(
                c.SYS.OSALNVItemInit.Req(
                    Id=nv_id,
                    ItemLen=len(value),
                    Value=t.ShortBytes(value[:244]),
                ),
                RspStatus=t.Status.NV_ITEM_UNINIT,
            )

        # 244 bytes is the most you can fit in a single `SYS.OSALNVWriteExt` command
        for offset in range(0, len(value), 244):
            await self.znp.request(
                c.SYS.OSALNVWriteExt.Req(
                    Id=nv_id,
                    Offset=offset,
                    Value=t.ShortBytes(value[offset : offset + 244]),
                ),
                RspStatus=t.Status.SUCCESS,
            )

    async def osal_read(self, nv_id: t.uint16_t, *, item_type):
        """
        Reads a complete value from NVRAM.

        Raises an `KeyError` error if the NVID doesn't exist.
        """

        # XXX: Some NVIDs don't really exist and Z-Stack behaves strangely with them
        if nv_id in PROXIED_NVIDS:
            read_rsp = await self.znp.request(
                c.SYS.OSALNVRead.Req(Id=nv_id, Offset=0),
                RspStatus=t.Status.SUCCESS,
            )

            value = self.deserialize(read_rsp.Value, item_type)
            LOGGER.debug('Read NVRAM["LEGACY"][0x%04x] = %r', nv_id, value)

            return value

        # Every item has a length, even missing ones
        length = (await self.znp.request(c.SYS.OSALNVLength.Req(Id=nv_id))).ItemLen

        if length == 0:
            raise KeyError(f"NV item does not exist: {nv_id!r}")

        data = b""

        try:
            while len(data) < length:
                read_rsp = await self.znp.request(
                    c.SYS.OSALNVReadExt.Req(Id=nv_id, Offset=len(data)),
                    RspStatus=t.Status.SUCCESS,
                )

                data += read_rsp.Value
        except InvalidCommandResponse as e:
            # Only expected status code is INVALID_PARAMETER
            assert e.response.Status == t.Status.INVALID_PARAMETER

            # Not all items can be read out due to security policies, though this can
            # easily be bypassed for some. The SAPI "ConfigId" is only 8 bits which
            # means some nvids are not able to read this way.
            if not self.znp.capabilities & t.MTCapabilities.SAPI or nv_id > 0xFF:
                raise SecurityError(
                    f"NV item cannot be read due to security constraints: {nv_id!r}"
                )

            read_rsp = await self.znp.request(
                c.SAPI.ZBReadConfiguration.Req(ConfigId=nv_id),
                RspStatus=t.Status.SUCCESS,
                RspConfigId=nv_id,
            )

            data = read_rsp.Value

        assert len(data) == length

        value = self.deserialize(data, item_type)
        LOGGER.debug('Read NVRAM["LEGACY"][0x%04x] = %r', nv_id, value)

        return value

    async def delete(
        self,
        *,
        sys_id: t.uint8_t = nvids.NvSysIds.ZSTACK,
        item_id: t.uint16_t,
        sub_id: t.uint16_t,
    ) -> bool:
        """
        Deletes a subitem from NVRAM. Returns whether or not the item existed.
        """

        delete_rsp = await self.znp.request(
            c.SYS.NVDelete.Req(SysId=sys_id, ItemId=item_id, SubId=sub_id)
        )

        return delete_rsp.Status == t.Status.SUCCESS

    async def write(
        self,
        *,
        sys_id: t.uint8_t = nvids.NvSysIds.ZSTACK,
        item_id: t.uint16_t,
        sub_id: t.uint16_t,
        value,
        create: bool = True,
    ) -> None:
        """
        Writes a value to NVRAM for the specified subsystem, item, and subitem.

        Calls to OSALNVWrite(sub_id=1) in newer Z-Stack releases are really calls to
        NVWrite(sys_id=ZSTACK, item_id=LEGACY, sub_id=1) in the background.
        """

        value = self.serialize(value)
        length = (
            await self.znp.request(
                c.SYS.NVLength.Req(SysId=sys_id, ItemId=item_id, SubId=sub_id)
            )
        ).Length

        if length != len(value) and not (
            sys_id == nvids.NvSysIds.ZSTACK
            and item_id in PROXIED_NVIDS
            and sub_id == 0x0000
        ):
            if not create:
                if length == 0:
                    raise KeyError(
                        f"NV item does not exist:"
                        f" sys_id={sys_id!r} item_id={item_id!r} sub_id={sub_id!r}"
                    )
                else:
                    raise ValueError(
                        f"Stored length and actual length differ:"
                        f" {length} != {len(value)}"
                    )

            if length != 0:
                await self.znp.request(
                    c.SYS.NVDelete.Req(SysId=sys_id, ItemId=item_id, SubId=sub_id),
                    RspStatus=t.Status.SUCCESS,
                )

            create_rsp = await self.znp.request(
                c.SYS.NVCreate.Req(
                    SysId=sys_id,
                    ItemId=item_id,
                    SubId=sub_id,
                    Length=len(value),
                )
            )

            if create_rsp.Status not in (t.Status.SUCCESS, t.Status.NV_ITEM_UNINIT):
                raise InvalidCommandResponse("Bad create status", create_rsp)

        # 244 bytes is the most you can fit in a single `SYS.NVWrite` command
        for offset in range(0, len(value), 244):
            await self.znp.request(
                c.SYS.NVWrite.Req(
                    SysId=sys_id,
                    ItemId=item_id,
                    SubId=sub_id,
                    Value=t.ShortBytes(value[offset : offset + 244]),
                    Offset=0,
                ),
                RspStatus=t.Status.SUCCESS,
            )

    async def read(
        self,
        *,
        sys_id: t.uint8_t = nvids.NvSysIds.ZSTACK,
        item_id: t.uint16_t,
        sub_id: t.uint16_t,
        item_type,
    ) -> bytes:
        """
        Reads a value from NVRAM for the specified subsystem, item, and subitem.

        Calls to OSALNVRead(sub_id=1) in newer Z-Stack releases are really calls to
        NVRead(sys_id=ZSTACK, item_id=LEGACY, sub_id=1) in the background.

        Raises an `KeyError` error if the NVID doesn't exist.
        """

        length_rsp = await self.znp.request(
            c.SYS.NVLength.Req(SysId=sys_id, ItemId=item_id, SubId=sub_id)
        )
        length = length_rsp.Length

        if length == 0:
            raise KeyError(
                f"NV item does not exist:"
                f" sys_id={sys_id!r} item_id={item_id!r} sub_id={sub_id!r}"
            )

        data = b""

        while len(data) < length:
            read_rsp = await self.znp.request(
                c.SYS.NVRead.Req(
                    SysId=sys_id,
                    ItemId=item_id,
                    SubId=sub_id,
                    Offset=len(data),
                    Length=length,
                ),
                RspStatus=t.Status.SUCCESS,
            )

            data += read_rsp.Value

        assert len(data) == length

        value = self.deserialize(data, item_type)
        LOGGER.debug("Read NVRAM[%s][%s][0x%04x] = %r", sys_id, item_id, sub_id, value)

        return value

    async def write_table(
        self,
        *,
        sys_id: t.uint8_t = nvids.NvSysIds.ZSTACK,
        item_id: t.uint16_t,
        values,
        fill_value,
    ) -> None:
        for sub_id, value in itertools.zip_longest(
            range(0x0000, 0xFFFF + 1), values, fillvalue=fill_value
        ):
            value = self.serialize(value)

            try:
                await self.write(
                    sys_id=sys_id,
                    item_id=item_id,
                    sub_id=sub_id,
                    value=value,
                    create=False,
                )
            except KeyError:
                break

    async def osal_write_table(
        self, start_nvid: t.uint16_t, end_nvid: t.uint16_t, values, *, fill_value
    ) -> None:
        values = list(values)

        for nvid, value in itertools.zip_longest(
            range(start_nvid, end_nvid + 1), values, fillvalue=fill_value
        ):
            value = self.serialize(value)

            try:
                await self.osal_write(
                    nv_id=nvid,
                    value=value,
                    create=False,
                )
            except KeyError:
                break

    async def read_table(
        self,
        *,
        sys_id: t.uint8_t = nvids.NvSysIds.ZSTACK,
        item_id: t.uint16_t,
        item_type=t.Bytes,
    ):
        for sub_id in range(0x0000, 0xFFFF + 1):
            try:
                yield await self.read(
                    sys_id=sys_id,
                    item_id=item_id,
                    sub_id=sub_id,
                    item_type=item_type,
                )
            except KeyError:
                break

    async def osal_read_table(
        self,
        start_nvid: t.uint16_t,
        end_nvid: t.uint16_t,
        item_type=t.Bytes,
    ):
        for nvid in range(start_nvid, end_nvid + 1):
            try:
                yield await self.osal_read(nvid, item_type=item_type)
            except KeyError:
                break



import logging

LOGGER = logging.getLogger(__name__)
_TRACE = 5


def _find_trace_level() -> int:
    if logging.getLevelName(_TRACE) != f"Level {_TRACE}":
        # If a level 5 exists, use it
        return _TRACE
    elif hasattr(logging, "TRACE") and logging.NOTSET < logging.TRACE < logging.DEBUG:
        # If a valid TRACE level exists that is between 0 and 10, use it
        return logging.TRACE
    elif LOGGER.level == logging.DEBUG:
        # If `zigpy_znp.logger` is explicitly passed `DEBUG` as a log level, enable
        # `TRACE` logging under the `DEBUG` level
        return logging.DEBUG
    else:
        # Otherwise, do not log
        return logging.NOTSET


TRACE = _find_trace_level()



from . import basic, named, cstruct


class NwkKeyDesc(cstruct.CStruct):
    KeySeqNum: basic.uint8_t
    Key: named.KeyData


class NwkState(basic.enum_uint8):
    NWK_INIT = 0
    NWK_JOINING_ORPHAN = 1
    NWK_DISC = 2
    NWK_JOINING = 3
    NWK_ENDDEVICE = 4
    PAN_CHNL_SELECTION = 5
    PAN_CHNL_VERIFY = 6
    PAN_STARTING = 7
    NWK_ROUTER = 8
    NWK_REJOINING = 9


class NIB(cstruct.CStruct):
    # NwkState is 16 bits on newer platforms but we cheat a little and make it 8, since
    # its max value is 8. However, we can't allow values like 0xFF08 to be serialized.
    _padding_byte = b"\x00"

    SequenceNum: basic.uint8_t
    PassiveAckTimeout: basic.uint8_t
    MaxBroadcastRetries: basic.uint8_t
    MaxChildren: basic.uint8_t
    MaxDepth: basic.uint8_t
    MaxRouters: basic.uint8_t
    dummyNeighborTable: basic.uint8_t
    BroadcastDeliveryTime: basic.uint8_t
    ReportConstantCost: basic.uint8_t
    RouteDiscRetries: basic.uint8_t
    dummyRoutingTable: basic.uint8_t
    SecureAllFrames: basic.uint8_t
    SecurityLevel: basic.uint8_t
    SymLink: basic.uint8_t
    CapabilityFlags: basic.uint8_t

    TransactionPersistenceTime: basic.uint16_t

    nwkProtocolVersion: basic.uint8_t
    RouteDiscoveryTime: basic.uint8_t
    RouteExpiryTime: basic.uint8_t

    nwkDevAddress: named.NWK

    nwkLogicalChannel: basic.uint8_t

    nwkCoordAddress: named.NWK
    nwkCoordExtAddress: named.EUI64
    nwkPanId: basic.uint16_t

    # XXX: this is really a uint16_t but we pad with zeroes so it works out in the end
    nwkState: NwkState
    channelList: named.Channels

    beaconOrder: basic.uint8_t
    superFrameOrder: basic.uint8_t
    scanDuration: basic.uint8_t
    battLifeExt: basic.uint8_t

    allocatedRouterAddresses: basic.uint32_t
    allocatedEndDeviceAddresses: basic.uint32_t

    nodeDepth: basic.uint8_t

    extendedPANID: named.EUI64

    nwkKeyLoaded: named.Bool

    spare1: NwkKeyDesc
    spare2: NwkKeyDesc

    spare3: basic.uint8_t
    spare4: basic.uint8_t

    nwkLinkStatusPeriod: basic.uint8_t
    nwkRouterAgeLimit: basic.uint8_t
    nwkUseMultiCast: named.Bool
    nwkIsConcentrator: named.Bool
    nwkConcentratorDiscoveryTime: basic.uint8_t
    nwkConcentratorRadius: basic.uint8_t
    nwkAllFresh: basic.uint8_t

    nwkManagerAddr: named.NWK
    nwkTotalTransmissions: basic.uint16_t
    nwkUpdateId: basic.uint8_t


class Beacon(cstruct.CStruct):
    """Beacon message."""

    Src: named.NWK
    PanId: named.PanId
    Channel: basic.uint8_t
    PermitJoining: basic.uint8_t
    RouterCapacity: basic.uint8_t
    DeviceCapacity: basic.uint8_t
    ProtocolVersion: basic.uint8_t
    StackProfile: basic.uint8_t
    LQI: basic.uint8_t
    Depth: basic.uint8_t
    UpdateId: basic.uint8_t
    ExtendedPanId: named.ExtendedPanId


class TCLinkKey(cstruct.CStruct):
    ExtAddr: named.EUI64
    Key: named.KeyData
    TxFrameCounter: basic.uint32_t
    RxFrameCounter: basic.uint32_t


class NwkActiveKeyItems(cstruct.CStruct):
    Active: NwkKeyDesc
    FrameCounter: basic.uint32_t


class KeyType(named.MissingEnumMixin, basic.enum_uint8):
    NONE = 0

    # Standard Network Key
    NWK = 1
    # Application Master Key
    APP_MASTER = 2
    # Application Link Key
    APP_LINK = 3
    # Trust Center Link Key
    TC_LINK = 4

    # XXX: just "6" in the Z-Stack source
    UNKNOWN_6 = 6


class KeyAttributes(basic.enum_uint8):
    # Used for IC derived keys
    PROVISIONAL_KEY = 0x00
    # Unique key that is not verified
    UNVERIFIED_KEY = 0x01
    # Unique key that got verified by ZC
    VERIFIED_KEY = 0x02

    # Internal definitions

    # Use default key to join
    DISTRIBUTED_DEFAULT_KEY = 0xFC
    # Joined a network which is not R21 nwk, so TCLK process finished.
    NON_R21_NWK_JOINED = 0xFD
    # Unique key that got verified by Joining device.
    # This means that key is stored as plain text (not seed hashed)
    VERIFIED_KEY_JOINING_DEV = 0xFE
    # Entry using default key
    DEFAULT_KEY = 0xFF


class TCLKDevEntry(cstruct.CStruct):
    _padding_byte = b"\x00"

    txFrmCntr: basic.uint32_t
    rxFrmCntr: basic.uint32_t

    extAddr: named.EUI64
    keyAttributes: KeyAttributes
    keyType: KeyType

    # For Unique key this is the number of shifts
    # for IC this is the offset on the NvId index
    SeedShift_IcIndex: basic.uint8_t


class NwkSecMaterialDesc(cstruct.CStruct):
    FrameCounter: basic.uint32_t
    ExtendedPanID: named.EUI64


class AddrMgrUserType(basic.enum_flag_uint8):
    Default = 0x00
    Assoc = 0x01
    Security = 0x02
    Binding = 0x04
    Private1 = 0x08


class AddrMgrEntry(cstruct.CStruct):
    type: AddrMgrUserType
    nwkAddr: named.NWK
    extAddr: named.EUI64


class AddressManagerTable(basic.CompleteList, item_type=AddrMgrEntry):
    pass


class AuthenticationOption(basic.enum_uint8):
    NotAuthenticated = 0x00
    AuthenticatedCBCK = 0x01
    AuthenticatedEA = 0x02


class APSKeyDataTableEntry(cstruct.CStruct):
    Key: named.KeyData
    TxFrameCounter: basic.uint32_t
    RxFrameCounter: basic.uint32_t


class APSLinkKeyTableEntry(cstruct.CStruct):
    AddressManagerIndex: basic.uint16_t
    LinkKeyNvId: basic.uint16_t
    AuthenticationState: AuthenticationOption


class APSLinkKeyTable(
    basic.LVList, length_type=basic.uint16_t, item_type=APSLinkKeyTableEntry
):
    pass



from __future__ import annotations

import os
import time
import asyncio
import logging
import itertools
import contextlib
import dataclasses
from collections import Counter, defaultdict

import zigpy.state
import async_timeout
import zigpy.zdo.types as zdo_t

import zigpy_znp.const as const
import zigpy_znp.types as t
import zigpy_znp.config as conf
import zigpy_znp.logger as log
import zigpy_znp.commands as c
from zigpy_znp import uart
from zigpy_znp.nvram import NVRAMHelper
from zigpy_znp.utils import (
    CatchAllResponse,
    BaseResponseListener,
    OneShotResponseListener,
    CallbackResponseListener,
)
from zigpy_znp.frames import GeneralFrame
from zigpy_znp.exceptions import CommandNotRecognized, InvalidCommandResponse
from zigpy_znp.types.nvids import ExNvIds, OsalNvIds

LOGGER = logging.getLogger(__name__)


# All of these are in seconds
AFTER_BOOTLOADER_SKIP_BYTE_DELAY = 2.5
NETWORK_COMMISSIONING_TIMEOUT = 30
BOOTLOADER_PIN_TOGGLE_DELAY = 0.15
CONNECT_PING_TIMEOUT = 0.50
CONNECT_PROBE_TIMEOUT = 10


class ZNP:
    def __init__(self, config: conf.ConfigType):
        self._uart = None
        self._app = None
        self._config = config

        self._listeners = defaultdict(list)
        self._sync_request_lock = asyncio.Lock()

        self.capabilities = None
        self.version = None

        self.nvram = NVRAMHelper(self)
        self.network_info: zigpy.state.NetworkInformation = None
        self.node_info: zigpy.state.NodeInfo = None

    def set_application(self, app):
        assert self._app is None
        self._app = app

    async def detect_zstack_version(self) -> float:
        """
        Feature detects the major version of Z-Stack running on the device.
        """

        # Z-Stack 1.2 does not have the AppConfig subsystem
        if not self.capabilities & t.MTCapabilities.APP_CNF:
            return 1.2

        try:
            # Only Z-Stack 3.30+ has the new NVRAM system
            await self.nvram.read(
                item_id=ExNvIds.TCLK_TABLE,
                sub_id=0x0000,
                item_type=t.Bytes,
            )
            return 3.30
        except KeyError:
            return 3.30
        except CommandNotRecognized:
            return 3.0

    async def load_network_info(self, *, load_devices=False):
        """
        Loads low-level network information from NVRAM.
        Loading key data greatly increases the runtime so it not enabled by default.
        """

        from zigpy_znp.znp import security

        is_on_network = None
        nib = None

        try:
            nib = await self.nvram.osal_read(OsalNvIds.NIB, item_type=t.NIB)
        except KeyError:
            is_on_network = False
        else:
            is_on_network = nib.nwkLogicalChannel != 0 and nib.nwkKeyLoaded

            if is_on_network and self.version >= 3.0:
                # This NVRAM item is the very first thing initialized in `zgInit`
                is_on_network = (
                    await self.nvram.osal_read(
                        OsalNvIds.BDBNODEISONANETWORK, item_type=t.uint8_t
                    )
                    == 1
                )

        if not is_on_network:
            raise ValueError("Device is not a part of a network")

        key_desc = await self.nvram.osal_read(
            OsalNvIds.NWK_ACTIVE_KEY_INFO, item_type=t.NwkKeyDesc
        )

        tc_frame_counter = await security.read_tc_frame_counter(
            self, ext_pan_id=nib.extendedPANID
        )

        network_info = zigpy.state.NetworkInformation(
            extended_pan_id=nib.extendedPANID,
            pan_id=nib.nwkPanId,
            nwk_update_id=nib.nwkUpdateId,
            channel=nib.nwkLogicalChannel,
            channel_mask=nib.channelList,
            security_level=nib.SecurityLevel,
            network_key=zigpy.state.Key(
                key=key_desc.Key,
                seq=key_desc.KeySeqNum,
                tx_counter=tc_frame_counter,
                rx_counter=None,
                partner_ieee=None,
            ),
            tc_link_key=None,
            children=[],
            nwk_addresses={},
            key_table=[],
            stack_specific=None,
        )

        if self.version > 1.2:
            tclk_seed = await self.nvram.osal_read(
                OsalNvIds.TCLK_SEED, item_type=t.KeyData
            )

            network_info.stack_specific = {
                "zstack": {
                    "tclk_seed": tclk_seed.serialize().hex(),
                }
            }

        # This takes a few seconds
        if load_devices:
            for dev in await security.read_devices(self):
                if dev.node_info.nwk == 0xFFFE:
                    dev = dev.replace(
                        node_info=dataclasses.replace(dev.node_info, nwk=None)
                    )

                if dev.is_child:
                    network_info.children.append(dev.node_info)
                elif dev.node_info.nwk is not None:
                    network_info.nwk_addresses[dev.node_info.ieee] = dev.node_info.nwk

                if dev.key is not None:
                    network_info.key_table.append(dev.key)

        ieee = await self.nvram.osal_read(OsalNvIds.EXTADDR, item_type=t.EUI64)
        logical_type = await self.nvram.osal_read(
            OsalNvIds.LOGICAL_TYPE, item_type=t.DeviceLogicalType
        )

        node_info = zigpy.state.NodeInfo(
            ieee=ieee,
            nwk=nib.nwkDevAddress,
            logical_type=zdo_t.LogicalType(logical_type),
        )

        self.network_info = network_info
        self.node_info = node_info

    async def write_network_info(
        self,
        *,
        network_info: zigpy.state.NetworkInformation,
        node_info: zigpy.state.NodeInfo,
    ) -> None:
        """
        Writes network and node state to NVRAM.
        """

        from zigpy_znp.znp import security

        # Delete any existing NV items that store formation state
        await self.nvram.osal_delete(OsalNvIds.HAS_CONFIGURED_ZSTACK1)
        await self.nvram.osal_delete(OsalNvIds.HAS_CONFIGURED_ZSTACK3)
        await self.nvram.osal_delete(OsalNvIds.ZIGPY_ZNP_MIGRATION_ID)
        await self.nvram.osal_delete(OsalNvIds.BDBNODEISONANETWORK)

        # Instruct Z-Stack to reset everything on the next boot
        await self.nvram.osal_write(
            OsalNvIds.STARTUP_OPTION,
            t.StartupOptions.ClearState | t.StartupOptions.ClearConfig,
        )

        await self.reset()

        # Form a network with completely random settings to get NVRAM to a known state
        for item, value in {
            OsalNvIds.PANID: t.uint16_t(0xFFFF),
            OsalNvIds.APS_USE_EXT_PANID: t.EUI64(os.urandom(8)),
            OsalNvIds.PRECFGKEY: os.urandom(16),
            # XXX: Z2M requires this item to be False
            OsalNvIds.PRECFGKEYS_ENABLE: t.Bool(False),
            # Z-Stack will scan all of thse channels during formation
            OsalNvIds.CHANLIST: const.STARTUP_CHANNELS,
        }.items():
            await self.nvram.osal_write(item, value, create=True)

        # Z-Stack 3+ ignores `CHANLIST`
        if self.version > 1.2:
            await self.request(
                c.AppConfig.BDBSetChannel.Req(
                    IsPrimary=True, Channel=const.STARTUP_CHANNELS
                ),
                RspStatus=t.Status.SUCCESS,
            )
            await self.request(
                c.AppConfig.BDBSetChannel.Req(
                    IsPrimary=False, Channel=t.Channels.NO_CHANNELS
                ),
                RspStatus=t.Status.SUCCESS,
            )

        # Both startup sequences end with the same callback
        started_as_coordinator = self.wait_for_response(
            c.ZDO.StateChangeInd.Callback(State=t.DeviceState.StartedAsCoordinator)
        )

        LOGGER.debug("Forming temporary network")

        # Handle the startup progress messages
        async with self.capture_responses(
            [
                c.ZDO.StateChangeInd.Callback(
                    State=t.DeviceState.StartingAsCoordinator
                ),
                c.AppConfig.BDBCommissioningNotification.Callback(
                    partial=True,
                    Status=c.app_config.BDBCommissioningStatus.InProgress,
                ),
            ]
        ):
            try:
                if self.version > 1.2:
                    # Z-Stack 3 uses the BDB subsystem
                    commissioning_rsp = await self.request_callback_rsp(
                        request=c.AppConfig.BDBStartCommissioning.Req(
                            Mode=c.app_config.BDBCommissioningMode.NwkFormation
                        ),
                        RspStatus=t.Status.SUCCESS,
                        callback=c.AppConfig.BDBCommissioningNotification.Callback(
                            partial=True,
                            RemainingModes=c.app_config.BDBCommissioningMode.NONE,
                        ),
                        timeout=NETWORK_COMMISSIONING_TIMEOUT,
                    )

                    if (
                        commissioning_rsp.Status
                        != c.app_config.BDBCommissioningStatus.Success
                    ):
                        raise RuntimeError(
                            f"Network formation failed: {commissioning_rsp}"
                        )
                else:
                    # In Z-Stack 1.2.2, StartupFromApp actually does what it says
                    await self.request(
                        c.ZDO.StartupFromApp.Req(StartDelay=100),
                        RspState=c.zdo.StartupState.NewNetworkState,
                    )

                # Both versions still end with this callback
                await started_as_coordinator
            except asyncio.TimeoutError as e:
                raise RuntimeError(
                    "Network formation refused, RF environment is likely too noisy."
                    " Temporarily unscrew the antenna or shield the coordinator"
                    " with metal until a network is formed."
                ) from e

        LOGGER.debug("Waiting for NIB to stabilize")

        # Even though the device says it is "ready" at this point, it takes a few more
        # seconds for `_NIB.nwkState` to switch from `NwkState.NWK_INIT`. There does
        # not appear to be any user-facing MT command to read this information.
        while True:
            try:
                nib = await self.nvram.osal_read(OsalNvIds.NIB, item_type=t.NIB)
            except KeyError:
                pass
            else:
                LOGGER.debug("Current NIB is %s", nib)

                # Usually this works after the first attempt
                if nib.nwkLogicalChannel != 0 and nib.nwkPanId != 0xFFFE:
                    break

            await asyncio.sleep(1)

        await self.reset()

        LOGGER.debug("Writing actual network settings")

        # Now that we have a formed network, update its state
        nib = await self.nvram.osal_read(OsalNvIds.NIB, item_type=t.NIB)
        nib = nib.replace(
            nwkDevAddress=node_info.nwk,
            nwkPanId=network_info.pan_id,
            extendedPANID=network_info.extended_pan_id,
            nwkUpdateId=network_info.nwk_update_id,
            nwkLogicalChannel=network_info.channel,
            channelList=network_info.channel_mask,
            SecurityLevel=network_info.security_level,
            nwkManagerAddr=network_info.nwk_manager_id,
            nwkCoordAddress=0x0000,
        )

        key_info = t.NwkActiveKeyItems(
            Active=t.NwkKeyDesc(
                KeySeqNum=network_info.network_key.seq,
                Key=network_info.network_key.key,
            ),
            FrameCounter=network_info.network_key.tx_counter,
        )

        nvram = {
            OsalNvIds.NIB: nib,
            OsalNvIds.APS_USE_EXT_PANID: network_info.extended_pan_id,
            OsalNvIds.EXTENDED_PAN_ID: network_info.extended_pan_id,
            OsalNvIds.PRECFGKEY: key_info.Active.Key,
            OsalNvIds.CHANLIST: network_info.channel_mask,
            OsalNvIds.EXTADDR: node_info.ieee,
            OsalNvIds.LOGICAL_TYPE: t.DeviceLogicalType(node_info.logical_type),
            OsalNvIds.NWK_ACTIVE_KEY_INFO: key_info.Active,
            OsalNvIds.NWK_ALTERN_KEY_INFO: const.EMPTY_KEY,
        }

        tclk_seed = None

        if self.version > 1.2:
            if (
                network_info.stack_specific is not None
                and network_info.stack_specific.get("zstack", {}).get("tclk_seed")
            ):
                tclk_seed, _ = t.KeyData.deserialize(
                    bytes.fromhex(network_info.stack_specific["zstack"]["tclk_seed"])
                )
            else:
                tclk_seed = t.KeyData(os.urandom(16))

            nvram[OsalNvIds.TCLK_SEED] = tclk_seed
        else:
            nvram[OsalNvIds.TCLK_SEED] = const.DEFAULT_TC_LINK_KEY

        for key, value in nvram.items():
            await self.nvram.osal_write(key, value, create=True)

        await security.write_tc_frame_counter(
            self,
            network_info.network_key.tx_counter,
            ext_pan_id=network_info.extended_pan_id,
        )

        devices = {}

        for child in network_info.children or []:
            devices[child.ieee] = security.StoredDevice(
                node_info=dataclasses.replace(
                    child, nwk=0xFFFE if child.nwk is None else child.nwk
                ),
                key=None,
                is_child=True,
            )

        for key in network_info.key_table or []:
            device = devices.get(key.partner_ieee)

            if device is None:
                device = security.StoredDevice(
                    node_info=zigpy.state.NodeInfo(
                        nwk=network_info.nwk_addresses.get(key.partner_ieee, 0xFFFE),
                        ieee=key.partner_ieee,
                        logical_type=None,
                    ),
                    key=None,
                    is_child=False,
                )

            devices[key.partner_ieee] = device.replace(key=key)

        LOGGER.debug("Writing children and keys")

        # Recompute the TCLK if necessary
        if self.version > 1.2:
            optimal_tclk_seed = security.find_optimal_tclk_seed(
                devices.values(), tclk_seed
            )

            if tclk_seed != optimal_tclk_seed:
                LOGGER.warning(
                    "Provided TCLK seed %s is not optimal, using %s instead.",
                    tclk_seed,
                    optimal_tclk_seed,
                )

                await self.nvram.osal_write(OsalNvIds.TCLK_SEED, optimal_tclk_seed)
                tclk_seed = optimal_tclk_seed

        await security.write_devices(
            znp=self,
            devices=list(devices.values()),
            tclk_seed=tclk_seed,
            counter_increment=0,
        )

        if self.version == 1.2:
            await self.nvram.osal_write(
                OsalNvIds.HAS_CONFIGURED_ZSTACK1,
                const.ZSTACK_CONFIGURE_SUCCESS,
                create=True,
            )
        else:
            await self.nvram.osal_write(
                OsalNvIds.HAS_CONFIGURED_ZSTACK3,
                const.ZSTACK_CONFIGURE_SUCCESS,
                create=True,
            )

        LOGGER.debug("Done!")

    async def reset(self) -> None:
        """
        Performs a soft reset within Z-Stack.
        A hard reset resets the serial port, causing the device to disconnect.
        """

        await self.request_callback_rsp(
            request=c.SYS.ResetReq.Req(Type=t.ResetType.Soft),
            callback=c.SYS.ResetInd.Callback(partial=True),
        )

    @property
    def _port_path(self) -> str:
        return self._config[conf.CONF_DEVICE][conf.CONF_DEVICE_PATH]

    @property
    def _znp_config(self) -> conf.ConfigType:
        return self._config[conf.CONF_ZNP_CONFIG]

    async def _skip_bootloader(self) -> c.SYS.Ping.Rsp:
        """
        Attempt to skip the bootloader and return the ping response.
        """

        async def ping_task():
            LOGGER.debug("Toggling RTS/DTR pins to skip bootloader or reset chip")

            # The default sequence is DTR=false and RTS toggling false/true/false
            for dtr, rts in zip(
                self._znp_config[conf.CONF_CONNECT_DTR_STATES],
                self._znp_config[conf.CONF_CONNECT_RTS_STATES],
            ):
                self._uart.set_dtr_rts(dtr=dtr, rts=rts)
                await asyncio.sleep(BOOTLOADER_PIN_TOGGLE_DELAY)

            # First, just try pinging
            try:
                async with async_timeout.timeout(CONNECT_PING_TIMEOUT):
                    return await self.request(c.SYS.Ping.Req())
            except asyncio.TimeoutError:
                pass

            # If that doesn't work, send the bootloader skip bytes and try again.
            # Sending a bunch at a time fixes UART issues when the radio was previously
            # probed with another library at a different baudrate.
            LOGGER.debug("Sending CC253x bootloader skip bytes")
            self._uart.write(256 * bytes([c.ubl.BootloaderRunMode.FORCE_RUN]))

            await asyncio.sleep(AFTER_BOOTLOADER_SKIP_BYTE_DELAY)

            # At this point we have nothing left to try
            while True:
                try:
                    async with async_timeout.timeout(2 * CONNECT_PING_TIMEOUT):
                        return await self.request(c.SYS.Ping.Req())
                except asyncio.TimeoutError:
                    pass

        async with self.capture_responses([CatchAllResponse()]) as responses:
            ping_task = asyncio.create_task(ping_task())

            try:
                async with async_timeout.timeout(CONNECT_PROBE_TIMEOUT):
                    result = await responses.get()
            except Exception:
                ping_task.cancel()
                raise
            else:
                LOGGER.debug("Radio is alive: %s", result)

            # Give the ping task a little bit extra time to finish. Radios often queue
            # requests so when the reset indication is received, they will all be
            # immediately answered
            if not ping_task.done():
                LOGGER.debug("Giving ping task %0.2fs to finish", CONNECT_PING_TIMEOUT)

                try:
                    async with async_timeout.timeout(CONNECT_PING_TIMEOUT):
                        result = await ping_task
                except asyncio.TimeoutError:
                    ping_task.cancel()

        if isinstance(result, c.SYS.Ping.Rsp):
            return result

        return await self.request(c.SYS.Ping.Req())

    async def connect(self, *, test_port=True) -> None:
        """
        Connects to the device specified by the "device" section of the config dict.

        The `test_port` kwarg allows port testing to be disabled, mainly to get into the
        bootloader.
        """

        # So we cannot connect twice
        assert self._uart is None

        try:
            self._uart = await uart.connect(self._config[conf.CONF_DEVICE], self)

            # To allow the ZNP interface to be used for bootloader commands, we have to
            # prevent any data from being sent
            if test_port:
                # The reset indication callback is sent when some sticks start up
                self.capabilities = (await self._skip_bootloader()).Capabilities

                # We need to know how structs are packed to deserialize frames correctly
                await self.nvram.determine_alignment()
                self.version = await self.detect_zstack_version()

                LOGGER.debug("Detected Z-Stack %s", self.version)
        except (Exception, asyncio.CancelledError):
            LOGGER.debug("Connection to %s failed, cleaning up", self._port_path)
            self.close()
            raise

        LOGGER.debug("Connected to %s at %s baud", self._uart.name, self._uart.baudrate)

    def connection_made(self) -> None:
        """
        Called by the UART object when a connection has been made.
        """
        pass

    def connection_lost(self, exc) -> None:
        """
        Called by the UART object to indicate that the port was closed. Propagates up
        to the `ControllerApplication` that owns this ZNP instance.
        """

        LOGGER.debug("We were disconnected from %s: %s", self._port_path, exc)

        if self._app is not None:
            self._app.connection_lost(exc)

    def close(self) -> None:
        """
        Cleans up resources, namely the listener queues.

        Calling this will reset ZNP to the same internal state as a fresh ZNP instance.
        """

        self._app = None

        for header, listeners in self._listeners.items():
            for listener in listeners:
                listener.cancel()

        self._listeners.clear()
        self.version = None
        self.capabilities = None

        if self._uart is not None:
            self._uart.close()
            self._uart = None

    def remove_listener(self, listener: BaseResponseListener) -> None:
        """
        Unbinds a listener from ZNP.

        Used by `wait_for_responses` to remove listeners for completed futures,
        regardless of their completion reason.
        """

        # If ZNP is closed while it's still running, `self._listeners` will be empty.
        if not self._listeners:
            return

        LOGGER.log(log.TRACE, "Removing listener %s", listener)

        for header in listener.matching_headers():
            try:
                self._listeners[header].remove(listener)
            except ValueError:
                pass

            if not self._listeners[header]:
                LOGGER.log(
                    log.TRACE, "Cleaning up empty listener list for header %s", header
                )
                del self._listeners[header]

        counts = Counter()

        for listener in itertools.chain.from_iterable(self._listeners.values()):
            counts[type(listener)] += 1

        LOGGER.log(
            log.TRACE,
            "There are %d callbacks and %d one-shot listeners remaining",
            counts[CallbackResponseListener],
            counts[OneShotResponseListener],
        )

    def frame_received(self, frame: GeneralFrame) -> bool:
        """
        Called when a frame has been received. Returns whether or not the frame was
        handled by any listener.

        XXX: Can be called multiple times in a single event loop step!
        """

        if frame.header not in c.COMMANDS_BY_ID:
            LOGGER.error("Received an unknown frame: %s", frame)
            return

        command_cls = c.COMMANDS_BY_ID[frame.header]

        try:
            command = command_cls.from_frame(frame, align=self.nvram.align_structs)
        except ValueError:
            # Some commands can be received corrupted. They are not useful:
            # https://github.com/home-assistant/core/issues/50005
            if command_cls == c.ZDO.ParentAnnceRsp.Callback:
                LOGGER.warning("Failed to parse broken %s as %s", frame, command_cls)
                return

            raise

        LOGGER.debug("Received command: %s", command)

        matched = False
        one_shot_matched = False

        for listener in (
            self._listeners[command.header] + self._listeners[CatchAllResponse.header]
        ):
            # XXX: A single response should *not* resolve multiple one-shot listeners!
            #      `future.add_done_callback` doesn't remove our listeners synchronously
            #      so doesn't prevent this from happening.
            if one_shot_matched and isinstance(listener, OneShotResponseListener):
                continue

            if not listener.resolve(command):
                LOGGER.log(log.TRACE, "%s does not match %s", command, listener)
                continue

            matched = True
            LOGGER.log(log.TRACE, "%s matches %s", command, listener)

            if isinstance(listener, OneShotResponseListener):
                one_shot_matched = True

        if not matched:
            self._unhandled_command(command)

        return matched

    def _unhandled_command(self, command: t.CommandBase):
        """
        Called when a command that is not handled by any listener is received.
        """

        LOGGER.warning("Received an unhandled command: %s", command)

    @contextlib.asynccontextmanager
    async def capture_responses(self, responses):
        """
        Captures all matched responses in a queue within the context manager.
        """

        queue = asyncio.Queue()
        listener = self.callback_for_responses(responses, callback=queue.put_nowait)

        try:
            yield queue
        finally:
            self.remove_listener(listener)

    def callback_for_responses(self, responses, callback) -> CallbackResponseListener:
        """
        Creates a callback listener that matches any of the provided responses.

        Only exists for consistency with `wait_for_responses`, since callbacks can be
        executed more than once.
        """

        listener = CallbackResponseListener(responses, callback=callback)

        LOGGER.log(log.TRACE, "Creating callback %s", listener)

        for header in listener.matching_headers():
            self._listeners[header].append(listener)

        return listener

    def callback_for_response(
        self, response: t.CommandBase, callback
    ) -> CallbackResponseListener:
        """
        Creates a callback listener for a single response.
        """

        return self.callback_for_responses([response], callback)

    def wait_for_responses(self, responses, *, context=False) -> asyncio.Future:
        """
        Creates a one-shot listener that matches any *one* of the given responses.
        """

        listener = OneShotResponseListener(responses)

        LOGGER.log(log.TRACE, "Creating one-shot listener %s", listener)

        for header in listener.matching_headers():
            self._listeners[header].append(listener)

        # Remove the listener when the future is done, not only when it gets a result
        listener.future.add_done_callback(lambda _: self.remove_listener(listener))

        if context:
            return listener.future, listener
        else:
            return listener.future

    def wait_for_response(self, response: t.CommandBase) -> asyncio.Future:
        """
        Creates a one-shot listener for a single response.
        """

        return self.wait_for_responses([response])

    async def request(self, request: t.CommandBase, **response_params) -> t.CommandBase:
        """
        Sends a SREQ/AREQ request and returns its SRSP (only for SREQ), failing if any
        of the SRSP's parameters don't match `response_params`.
        """

        # Common mistake is to do `znp.request(c.SYS.Ping())`
        if type(request) is not request.Req:
            raise ValueError(f"Cannot send a command that isn't a request: {request!r}")

        # Construct a partial response out of the `Rsp*` kwargs if one is provided
        if request.Rsp:
            renamed_response_params = {}

            for param, value in response_params.items():
                if not param.startswith("Rsp"):
                    raise KeyError(
                        f"All response params must start with 'Rsp': {param!r}"
                    )

                renamed_response_params[param.replace("Rsp", "", 1)] = value

            # Construct our response before we send the request so that we fail early
            partial_response = request.Rsp(partial=True, **renamed_response_params)
        elif response_params:
            raise ValueError(
                f"Command has no response so response_params={response_params} "
                f"will have no effect"
            )

        frame = request.to_frame(align=self.nvram.align_structs)

        # We should only be sending one SREQ at a time, according to the spec
        async with self._sync_request_lock:
            LOGGER.debug("Sending request: %s", request)

            # If our request has no response, we cannot wait for one
            if not request.Rsp:
                LOGGER.debug("Request has no response, not waiting for one.")
                self._uart.send(frame)
                return

            # We need to create the response listener before we send the request
            response_future = self.wait_for_responses(
                [
                    request.Rsp(partial=True),
                    c.RPCError.CommandNotRecognized.Rsp(
                        partial=True, RequestHeader=request.header
                    ),
                ]
            )

            self._uart.send(frame)

            # We should get a SRSP in a reasonable amount of time
            async with async_timeout.timeout(self._znp_config[conf.CONF_SREQ_TIMEOUT]):
                # We lock until either a sync response is seen or an error occurs
                response = await response_future

        if isinstance(response, c.RPCError.CommandNotRecognized.Rsp):
            raise CommandNotRecognized(
                f"Fatal request error {response} in response to {request}"
            )

        # If the sync response we got is not what we wanted, this is an error
        if not partial_response.matches(response):
            raise InvalidCommandResponse(
                f"Expected SRSP response {partial_response}, got {response}", response
            )

        return response

    async def request_callback_rsp(
        self, *, request, callback, timeout=None, background=False, **response_params
    ):
        """
        Sends an SREQ, gets its SRSP confirmation, and waits for its real AREQ response.
        A bug-free version of:

            req_rsp = await req
            callback_rsp = await req_callback

        This is necessary because the SRSP and the AREQ may arrive in the same "chunk"
        from the UART and be handled in the same event loop step by ZNP.
        """

        # Every request should have a timeout to prevent deadlocks
        if timeout is None:
            timeout = self._znp_config[conf.CONF_ARSP_TIMEOUT]

        callback_rsp, listener = self.wait_for_responses([callback], context=True)

        # Typical request/response/callbacks are not backgrounded
        if not background:
            try:
                async with async_timeout.timeout(timeout):
                    await self.request(request, **response_params)

                    return await callback_rsp
            finally:
                self.remove_listener(listener)

        # Backgrounded callback handlers need to respect the provided timeout
        start_time = time.monotonic()

        try:
            async with async_timeout.timeout(timeout):
                request_rsp = await self.request(request, **response_params)
        except Exception:
            # If the SREQ/SRSP pair fails, we must cancel the AREQ listener
            self.remove_listener(listener)
            raise

        # If it succeeds, create a background task to receive the AREQ but take into
        # account the time it took to start the SREQ to ensure we do not grossly exceed
        # the timeout
        async def callback_catcher(timeout):
            try:
                async with async_timeout.timeout(timeout):
                    await callback_rsp
            finally:
                self.remove_listener(listener)

        callback_timeout = max(0, timeout - (time.monotonic() - start_time))
        asyncio.create_task(callback_catcher(callback_timeout))

        return request_rsp



import zigpy_znp.types as t


class BaseNvIds(t.enum_uint16):
    pass


class ZclPortNvIds(BaseNvIds):
    # ZCL Port NV IDs (Application Layer NV Items)
    SCENE_TABLE = 0x0001
    PROXY_TABLE = 0x0002
    SINK_TABLE = 0x0003


class NvSysIds(t.enum_uint8):
    NVDRVR = 0  # Refrain from use
    ZSTACK = 1
    TIMAC = 2
    REMOTI = 3
    BLE = 4
    _6MESH = 5
    TIOP = 6
    APP = 7


class ExNvIds(BaseNvIds):
    # OSAL NV Item IDs
    LEGACY = 0x0000
    ADDRMGR = 0x0001
    BINDING_TABLE = 0x0002
    DEVICE_LIST = 0x0003
    TCLK_TABLE = 0x0004
    TCLK_IC_TABLE = 0x0005
    APS_KEY_DATA_TABLE = 0x0006
    NWK_SEC_MATERIAL_TABLE = 0x0007


class OsalNvIds(BaseNvIds):
    # Introduced by zigbeer/zigbee-shepherd and now used by Zigbee2MQTT
    HAS_CONFIGURED_ZSTACK1 = 0x0F00
    HAS_CONFIGURED_ZSTACK3 = 0x0060

    # Although the docs say "IDs reserved for applications range from 0x0401 to 0x0FFF",
    # no OSAL NVID beyond 0x03FF is writable with the MT interface when using Z-Stack 3.
    ZIGPY_ZNP_MIGRATION_ID = 0x005F

    # OSAL NV item IDs
    EXTADDR = 0x0001
    BOOTCOUNTER = 0x0002
    STARTUP_OPTION = 0x0003
    START_DELAY = 0x0004

    # NWK Layer NV item IDs
    NIB = 0x0021
    DEVICE_LIST = 0x0022
    ADDRMGR = 0x0023
    POLL_RATE_OLD16 = 0x0024  # Deprecated when poll rate changed from 16 to 32 bits
    QUEUED_POLL_RATE = 0x0025
    RESPONSE_POLL_RATE = 0x0026
    REJOIN_POLL_RATE = 0x0027
    DATA_RETRIES = 0x0028
    POLL_FAILURE_RETRIES = 0x0029
    STACK_PROFILE = 0x002A
    INDIRECT_MSG_TIMEOUT = 0x002B
    ROUTE_EXPIRY_TIME = 0x002C
    EXTENDED_PAN_ID = 0x002D
    BCAST_RETRIES = 0x002E
    PASSIVE_ACK_TIMEOUT = 0x002F
    BCAST_DELIVERY_TIME = 0x0030
    NWK_MODE = 0x0031  # Deprecated, as this will always be Mesh
    CONCENTRATOR_ENABLE = 0x0032
    CONCENTRATOR_DISCOVERY = 0x0033
    CONCENTRATOR_RADIUS = 0x0034
    POLL_RATE = 0x0035
    CONCENTRATOR_RC = 0x0036
    NWK_MGR_MODE = 0x0037
    SRC_RTG_EXPIRY_TIME = 0x0038
    ROUTE_DISCOVERY_TIME = 0x0039
    NWK_ACTIVE_KEY_INFO = 0x003A
    NWK_ALTERN_KEY_INFO = 0x003B
    ROUTER_OFF_ASSOC_CLEANUP = 0x003C
    NWK_LEAVE_REQ_ALLOWED = 0x003D
    NWK_CHILD_AGE_ENABLE = 0x003E
    DEVICE_LIST_KA_TIMEOUT = 0x003F

    # APS Layer NV item IDs
    BINDING_TABLE = 0x0041
    GROUP_TABLE = 0x0042
    APS_FRAME_RETRIES = 0x0043
    APS_ACK_WAIT_DURATION = 0x0044
    APS_ACK_WAIT_MULTIPLIER = 0x0045
    BINDING_TIME = 0x0046
    APS_USE_EXT_PANID = 0x0047
    APS_USE_INSECURE_JOIN = 0x0048
    COMMISSIONED_NWK_ADDR = 0x0049

    APS_NONMEMBER_RADIUS = 0x004B  # Multicast non_member radius
    APS_LINK_KEY_TABLE = 0x004C
    APS_DUPREJ_TIMEOUT_INC = 0x004D
    APS_DUPREJ_TIMEOUT_COUNT = 0x004E
    APS_DUPREJ_TABLE_SIZE = 0x004F

    # System statistics and metrics NV ID
    DIAGNOSTIC_STATS = 0x0050

    # Additional NWK Layer NV item IDs
    NWK_PARENT_INFO = 0x0051
    NWK_ENDDEV_TIMEOUT_DEF = 0x0052
    END_DEV_TIMEOUT_VALUE = 0x0053
    END_DEV_CONFIGURATION = 0x0054

    BDBNODEISONANETWORK = 0x0055  # bdbNodeIsOnANetwork attribute
    BDBREPORTINGCONFIG = 0x0056

    # Security NV Item IDs
    SECURITY_LEVEL = 0x0061
    PRECFGKEY = 0x0062
    PRECFGKEYS_ENABLE = 0x0063
    # Deprecated Item as there is only one security mode supported now Z3.0
    SECURITY_MODE = 0x0064
    SECURE_PERMIT_JOIN = 0x0065
    APS_LINK_KEY_TYPE = 0x0066
    APS_ALLOW_R19_SECURITY = 0x0067
    DISTRIBUTED_KEY = 0x0068  # Default distributed nwk key Id. Nv ID not in use

    IMPLICIT_CERTIFICATE = 0x0069
    DEVICE_PRIVATE_KEY = 0x006A
    CA_PUBLIC_KEY = 0x006B
    KE_MAX_DEVICES = 0x006C

    USE_DEFAULT_TCLK = 0x006D
    # deprecated: TRUSTCENTER_ADDR (16-bit)   0x006E
    RNG_COUNTER = 0x006F
    RANDOM_SEED = 0x0070
    TRUSTCENTER_ADDR = 0x0071

    CERT_283 = 0x0072
    PRIVATE_KEY_283 = 0x0073
    PUBLIC_KEY_283 = 0x0074

    LEGACY_NWK_SEC_MATERIAL_TABLE_START = 0x0075
    LEGACY_NWK_SEC_MATERIAL_TABLE_END = 0x0080

    # ZDO NV Item IDs
    USERDESC = 0x0081
    NWKKEY = 0x0082
    PANID = 0x0083
    CHANLIST = 0x0084
    LEAVE_CTRL = 0x0085
    SCAN_DURATION = 0x0086
    LOGICAL_TYPE = 0x0087
    NWKMGR_MIN_TX = 0x0088
    NWKMGR_ADDR = 0x0089

    ZDO_DIRECT_CB = 0x008F

    # ZCL NV item IDs
    SCENE_TABLE = 0x0091
    MIN_FREE_NWK_ADDR = 0x0092
    MAX_FREE_NWK_ADDR = 0x0093
    MIN_FREE_GRP_ID = 0x0094
    MAX_FREE_GRP_ID = 0x0095
    MIN_GRP_IDS = 0x0096
    MAX_GRP_IDS = 0x0097
    OTA_BLOCK_REQ_DELAY = 0x0098

    # Non-standard NV item IDs
    SAPI_ENDPOINT = 0x00A1

    # NV Items Reserved for Commissioning Cluster Startup Attribute Set (SAS):
    # 0x00B1 - 0x00BF: Parameters related to APS and NWK layers
    SAS_SHORT_ADDR = 0x00B1
    SAS_EXT_PANID = 0x00B2
    SAS_PANID = 0x00B3
    SAS_CHANNEL_MASK = 0x00B4
    SAS_PROTOCOL_VER = 0x00B5
    SAS_STACK_PROFILE = 0x00B6
    SAS_STARTUP_CTRL = 0x00B7

    # 0x00C1 - 0x00CF: Parameters related to Security
    SAS_TC_ADDR = 0x00C1
    SAS_TC_MASTER_KEY = 0x00C2
    SAS_NWK_KEY = 0x00C3
    SAS_USE_INSEC_JOIN = 0x00C4
    SAS_PRECFG_LINK_KEY = 0x00C5
    SAS_NWK_KEY_SEQ_NUM = 0x00C6
    SAS_NWK_KEY_TYPE = 0x00C7
    SAS_NWK_MGR_ADDR = 0x00C8

    # 0x00D1 - 0x00DF: Current key parameters
    SAS_CURR_TC_MASTER_KEY = 0x00D1
    SAS_CURR_NWK_KEY = 0x00D2
    SAS_CURR_PRECFG_LINK_KEY = 0x00D3

    USE_NVOCMP = 0x00FF

    # NV Items Reserved for Trust Center Link Key Table entries
    # 0x0101 - 0x01FF
    TCLK_SEED = 0x0101  # Seed
    TCLK_JOIN_DEV = (
        0x0102  # Nv Id where Joining device store their APS key. Key is in plain text.
    )
    TCLK_DEFAULT = 0x0103  # Not accually a Nv Item but Id used by SecMgr

    LEGACY_TCLK_IC_TABLE_START = 0x0104  # Deprecated. Refer to EX_TCLK_IC_TABLE
    LEGACY_TCLK_IC_TABLE_END = 0x0110  # IC keys, referred with shift byte

    LEGACY_TCLK_TABLE_START = 0x0111  # Deprecated. Refer to EX_TCLK_TABLE
    LEGACY_TCLK_TABLE_END = 0x01FF

    # NV Items Reserved for APS Link Key Table entries
    # 0x0201 - 0x02FF
    LEGACY_APS_LINK_KEY_DATA_START = 0x0201  # Deprecated. Refer to EX_APS_KEY_TABLE
    LEGACY_APS_LINK_KEY_DATA_END = 0x02FF

    # NV items used to duplicate system elements
    DUPLICATE_BINDING_TABLE = 0x0300
    DUPLICATE_DEVICE_LIST = 0x0301
    DUPLICATE_DEVICE_LIST_KA_TIMEOUT = 0x0302

    # NV Items Reserved for Proxy Table entries
    # 0x0310 - 0x031F
    LEGACY_PROXY_TABLE_START = 0x0310  # Deprecated. Refer to EX_GP_PROXY_TABLE
    LEGACY_PROXY_TABLE_END = 0x031F

    # NV Items Reserved for Sink Table entries
    # 0x0320 - 0x032F
    LEGACY_SINK_TABLE_START = 0x0320  # Deprecated. Refer to EX_GP_SINK_TABLE
    LEGACY_SINK_TABLE_END = 0x032F

    APP_ITEM_1 = 0x0F01
    APP_ITEM_2 = 0x0F02
    APP_ITEM_3 = 0x0F03
    APP_ITEM_4 = 0x0F04
    APP_ITEM_5 = 0x0F05
    APP_ITEM_6 = 0x0F06

    RF_TEST_PARMS = 0x0F07

    UNKNOWN = 0x0F08

    INVALID_INDEX = 0xFFFF


NWK_NVID_TABLES = {
    OsalNvIds.LEGACY_NWK_SEC_MATERIAL_TABLE_START: (
        OsalNvIds.LEGACY_NWK_SEC_MATERIAL_TABLE_END
    ),
    OsalNvIds.LEGACY_TCLK_IC_TABLE_START: OsalNvIds.LEGACY_TCLK_IC_TABLE_END,
    OsalNvIds.LEGACY_TCLK_TABLE_START: OsalNvIds.LEGACY_TCLK_TABLE_END,
    OsalNvIds.LEGACY_APS_LINK_KEY_DATA_START: OsalNvIds.LEGACY_APS_LINK_KEY_DATA_END,
    OsalNvIds.LEGACY_PROXY_TABLE_START: OsalNvIds.LEGACY_PROXY_TABLE_END,
    OsalNvIds.LEGACY_SINK_TABLE_START: OsalNvIds.LEGACY_SINK_TABLE_END,
}

NWK_NVID_TABLE_KEYS = set(NWK_NVID_TABLES.keys()) | set(NWK_NVID_TABLES.values())


def is_secure_nvid(nvid: OsalNvIds) -> bool:
    """
    Returns whether or not an nvid may be prevented from being read from NVRAM.
    """

    if nvid in (
        OsalNvIds.IMPLICIT_CERTIFICATE,
        OsalNvIds.CA_PUBLIC_KEY,
        OsalNvIds.DEVICE_PRIVATE_KEY,
        OsalNvIds.NWK_ACTIVE_KEY_INFO,
        OsalNvIds.NWK_ALTERN_KEY_INFO,
        OsalNvIds.PRECFGKEY,
        OsalNvIds.TCLK_SEED,
    ):
        return True

    if OsalNvIds.LEGACY_TCLK_TABLE_START <= nvid <= OsalNvIds.LEGACY_TCLK_TABLE_END:
        return True

    if (
        OsalNvIds.LEGACY_APS_LINK_KEY_DATA_START
        <= nvid
        <= OsalNvIds.LEGACY_APS_LINK_KEY_DATA_END
    ):
        return True

    return False



import sys
import asyncio
import logging
import itertools
from collections import deque, defaultdict

import zigpy.zdo.types as zdo_t

import zigpy_znp.types as t
from zigpy_znp.tools.common import setup_parser
from zigpy_znp.zigbee.application import ControllerApplication

LOGGER = logging.getLogger(__name__)


async def perform_energy_scan(radio_path, num_scans=None):
    LOGGER.info("Starting up zigpy-znp")

    config = ControllerApplication.SCHEMA({"device": {"path": radio_path}})
    app = ControllerApplication(config)

    try:
        await app.startup(read_only=True)
    except RuntimeError as e:
        LOGGER.error("Could not start application: %s", e)
        LOGGER.error("Form a network with `python -m zigpy_znp.tools.form_network`")
        return

    LOGGER.info("Running scan...")

    # We compute an average over the last 5 scans
    channel_energies = defaultdict(lambda: deque([], maxlen=5))

    for i in itertools.count(start=1):
        if num_scans is not None and i > num_scans:
            break

        rsp = await app.get_device(nwk=0x0000).zdo.Mgmt_NWK_Update_req(
            zdo_t.NwkUpdate(
                ScanChannels=t.Channels.ALL_CHANNELS,
                ScanDuration=0x02,
                ScanCount=1,
            )
        )

        _, scanned_channels, _, _, energy_values = rsp

        for channel, energy in zip(scanned_channels, energy_values):
            energies = channel_energies[channel]
            energies.append(energy)

        total = 0xFF * len(energies)

        print(f"Channel energy (mean of {len(energies)} / {energies.maxlen}):")
        print("------------------------------------------------")
        print(" + Lower energy is better")
        print(" + Active Zigbee networks on a channel may still cause congestion")
        print(" + TX on 26 in North America may be with lower power due to regulations")
        print(" + Zigbee channels 15, 20, 25 fall between WiFi channels 1, 6, 11")
        print(" + Some Zigbee devices only join networks on channels 15, 20, and 25")
        print("------------------------------------------------")

        for channel, energies in channel_energies.items():
            count = sum(energies)
            asterisk = "*" if channel == 26 else " "

            print(
                f" - {channel:>02}{asterisk}  {count / total:>7.2%}  "
                + "#" * int(100 * count / total)
            )

        print()

    await app.shutdown()


async def main(argv):
    parser = setup_parser("Perform an energy scan")
    parser.add_argument(
        "-n",
        "--num-scans",
        dest="num_scans",
        type=int,
        default=None,
        help="Number of scans to perform before exiting",
    )

    args = parser.parse_args(argv)

    await perform_energy_scan(args.serial, num_scans=args.num_scans)


if __name__ == "__main__":
    asyncio.run(main(sys.argv[1:]))  # pragma: no cover



from __future__ import annotations

import os
import time
import random
import asyncio
import logging
import itertools
import contextlib

import zigpy.zcl
import zigpy.zdo
import zigpy.util
import zigpy.state
import zigpy.types
import zigpy.config
import zigpy.device
import async_timeout
import zigpy.endpoint
import zigpy.profiles
import zigpy.zdo.types as zdo_t
import zigpy.application
from zigpy.zcl import clusters
from zigpy.types import ExtendedPanId, deserialize as list_deserialize
from zigpy.zdo.types import CLUSTERS as ZDO_CLUSTERS, ZDOCmd, ZDOHeader, MultiAddress
from zigpy.exceptions import DeliveryError

import zigpy_znp.const as const
import zigpy_znp.types as t
import zigpy_znp.config as conf
import zigpy_znp.commands as c
from zigpy_znp.api import ZNP
from zigpy_znp.znp import security
from zigpy_znp.utils import combine_concurrent_calls
from zigpy_znp.exceptions import CommandNotRecognized, InvalidCommandResponse
from zigpy_znp.types.nvids import OsalNvIds
from zigpy_znp.zigbee.zdo_converters import ZDO_CONVERTERS

ZDO_ENDPOINT = 0
ZHA_ENDPOINT = 1
ZLL_ENDPOINT = 2

# All of these are in seconds
PROBE_TIMEOUT = 5
STARTUP_TIMEOUT = 5
ZDO_REQUEST_TIMEOUT = 15
DATA_CONFIRM_TIMEOUT = 8
DEVICE_JOIN_MAX_DELAY = 5
WATCHDOG_PERIOD = 30
BROADCAST_SEND_WAIT_DURATION = 3
MULTICAST_SEND_WAIT_DURATION = 3

REQUEST_MAX_RETRIES = 5
REQUEST_ERROR_RETRY_DELAY = 0.5

# Errors that go away on their own after waiting for a bit
REQUEST_TRANSIENT_ERRORS = {
    t.Status.BUFFER_FULL,
    t.Status.MAC_CHANNEL_ACCESS_FAILURE,
    t.Status.MAC_TRANSACTION_OVERFLOW,
    t.Status.MAC_NO_RESOURCES,
    t.Status.MEM_ERROR,
    t.Status.NWK_TABLE_FULL,
}

REQUEST_ROUTING_ERRORS = {
    t.Status.APS_NO_ACK,
    t.Status.APS_NOT_AUTHENTICATED,
    t.Status.NWK_NO_ROUTE,
    t.Status.NWK_INVALID_REQUEST,
    t.Status.MAC_NO_ACK,
    t.Status.MAC_TRANSACTION_EXPIRED,
}

REQUEST_RETRYABLE_ERRORS = REQUEST_TRANSIENT_ERRORS | REQUEST_ROUTING_ERRORS

LOGGER = logging.getLogger(__name__)


class ZNPCoordinator(zigpy.device.Device):
    """
    Coordinator zigpy device that keeps track of our endpoints and clusters.
    """

    @property
    def manufacturer(self):
        return "Texas Instruments"

    @property
    def model(self):
        if self.application._znp.version > 3.0:
            model = "CC1352/CC2652"
            version = "3.30+"
        else:
            model = "CC2538" if self.application._znp.nvram.align_structs else "CC2531"
            version = "Home 1.2" if self.application._znp.version == 1.2 else "3.0.x"

        return f"{model}, Z-Stack {version} (build {self.application._zstack_build_id})"


class ControllerApplication(zigpy.application.ControllerApplication):
    SCHEMA = conf.CONFIG_SCHEMA
    SCHEMA_DEVICE = conf.SCHEMA_DEVICE

    def __init__(self, config: conf.ConfigType):
        super().__init__(config=conf.CONFIG_SCHEMA(config))

        self._znp: ZNP | None = None

        # It's simpler to work with Task objects if they're never actually None
        self._reconnect_task = asyncio.Future()
        self._reconnect_task.cancel()

        self._watchdog_task = asyncio.Future()
        self._watchdog_task.cancel()

        self._version_rsp = None
        self._concurrent_requests_semaphore = None
        self._currently_waiting_requests = 0

        self._join_announce_tasks = {}

    ##################################################################
    # Implementation of the core zigpy ControllerApplication methods #
    ##################################################################

    @classmethod
    async def probe(cls, device_config: conf.ConfigType) -> bool:
        """
        Checks whether the device represented by `device_config` is a valid ZNP radio.
        Doesn't throw any errors.
        """

        znp = ZNP(conf.CONFIG_SCHEMA({conf.CONF_DEVICE: device_config}))
        LOGGER.debug("Probing %s", znp._port_path)

        try:
            # `ZNP.connect` times out on its own
            await znp.connect()
        except Exception as e:
            LOGGER.debug(
                "Failed to probe ZNP radio with config %s", device_config, exc_info=e
            )
            return False
        else:
            return True
        finally:
            znp.close()

    async def shutdown(self):
        """
        Gracefully shuts down the application and cleans up all resources.
        """

        self.close()

    def close(self):
        self._reconnect_task.cancel()
        self._watchdog_task.cancel()

        # This will close the UART, which will then close the transport
        if self._znp is not None:
            self._znp.close()
            self._znp = None

    async def startup(self, auto_form=False, force_form=False, read_only=False):
        """
        Performs application startup.

        This entails creating the ZNP object, connecting to the radio, potentially
        forming a network, and configuring our settings.
        """

        try:
            return await self._startup(
                auto_form=auto_form,
                force_form=force_form,
                read_only=read_only,
            )
        except Exception:
            await self.shutdown()
            raise

    async def _startup(self, auto_form=False, force_form=False, read_only=False):
        assert self._znp is None

        znp = ZNP(self.config)
        await znp.connect()

        # We only assign `self._znp` after it has successfully connected
        self._znp = znp
        self._znp.set_application(self)

        if not read_only and not force_form:
            await self._migrate_nvram()

        self._bind_callbacks()

        # Next, read out the NVRAM item that Zigbee2MQTT writes when it has configured
        # a device to make sure that our network settings will not be reset.
        if self._znp.version == 1.2:
            configured_nv_item = OsalNvIds.HAS_CONFIGURED_ZSTACK1
        else:
            configured_nv_item = OsalNvIds.HAS_CONFIGURED_ZSTACK3

        try:
            configured_value = await self._znp.nvram.osal_read(
                configured_nv_item, item_type=t.uint8_t
            )
        except KeyError:
            is_configured = False
        else:
            is_configured = configured_value == const.ZSTACK_CONFIGURE_SUCCESS

        if force_form:
            LOGGER.info("Forming a new network")
            await self.form_network()
        elif not is_configured:
            if not auto_form:
                raise RuntimeError("Cannot start application, network is not formed")
            elif read_only:
                raise RuntimeError(
                    "Cannot start application, network is not formed and read-only"
                )

            LOGGER.info("ZNP is not configured, forming a new network")

            # Network formation requires multiple resets so it will write the NVRAM
            # settings itself
            await self.form_network()
        else:
            # Issue a reset first to make sure we aren't permitting joins
            await self._znp.reset()

            LOGGER.info("ZNP is already configured, not forming a new network")

            if not read_only:
                await self._write_stack_settings(reset_if_changed=True)

        # At this point the device state should the same, regardless of whether we just
        # formed a new network or are restoring one
        if self.znp_config[conf.CONF_TX_POWER] is not None:
            await self.set_tx_power(dbm=self.znp_config[conf.CONF_TX_POWER])

        # Both versions of Z-Stack use this callback
        started_as_coordinator = self._znp.wait_for_response(
            c.ZDO.StateChangeInd.Callback(State=t.DeviceState.StartedAsCoordinator)
        )

        if self._znp.version == 1.2:
            # Z-Stack Home 1.2 has a simple startup sequence
            await self._znp.request(
                c.ZDO.StartupFromApp.Req(StartDelay=100),
                RspState=c.zdo.StartupState.RestoredNetworkState,
            )
        else:
            # Z-Stack 3 uses the BDB subsystem
            bdb_commissioning_done = self._znp.wait_for_response(
                c.AppConfig.BDBCommissioningNotification.Callback(
                    partial=True, RemainingModes=c.app_config.BDBCommissioningMode.NONE
                )
            )

            # According to the forums, this is the correct startup sequence, including
            # the formation failure error
            await self._znp.request_callback_rsp(
                request=c.AppConfig.BDBStartCommissioning.Req(
                    Mode=c.app_config.BDBCommissioningMode.NwkFormation
                ),
                RspStatus=t.Status.SUCCESS,
                callback=c.AppConfig.BDBCommissioningNotification.Callback(
                    partial=True,
                    Status=c.app_config.BDBCommissioningStatus.NetworkRestored,
                ),
            )

            await bdb_commissioning_done

        # The startup sequence should not take forever
        async with async_timeout.timeout(STARTUP_TIMEOUT):
            await started_as_coordinator

        self._version_rsp = await self._znp.request(c.SYS.Version.Req())

        # XXX: The CC2531 running Z-Stack Home 1.2 permanently permits joins on startup
        # unless they are explicitly disabled. We can't fix this but we can disable them
        # as early as possible to shrink the window of opportunity for unwanted joins.
        await self.permit(time_s=0)

        # The CC2531 running Z-Stack Home 1.2 overrides the LED setting if it is changed
        # before the coordinator has started.
        if self.znp_config[conf.CONF_LED_MODE] is not None:
            await self._set_led_mode(led=0xFF, mode=self.znp_config[conf.CONF_LED_MODE])

        await self.load_network_info()
        await self._register_endpoints()

        # Setup the coordinator as a zigpy device and initialize it to request node info
        self.devices[self.ieee] = ZNPCoordinator(self, self.ieee, self.nwk)
        await self.zigpy_device.schedule_initialize()

        # Now that we know what device we are, set the max concurrent requests
        if self.znp_config[conf.CONF_MAX_CONCURRENT_REQUESTS] == "auto":
            max_concurrent_requests = 16 if self._znp.nvram.align_structs else 2
        else:
            max_concurrent_requests = self.znp_config[conf.CONF_MAX_CONCURRENT_REQUESTS]

        self._concurrent_requests_semaphore = asyncio.Semaphore(max_concurrent_requests)

        LOGGER.info("Network settings")
        LOGGER.info("  Model: %s", self.zigpy_device.model)
        LOGGER.info("  Z-Stack version: %s", self._znp.version)
        LOGGER.info("  Z-Stack build id: %s", self._zstack_build_id)
        LOGGER.info("  Max concurrent requests: %s", max_concurrent_requests)
        LOGGER.info("  Channel: %s", self.channel)
        LOGGER.info("  PAN ID: 0x%04X", self.pan_id)
        LOGGER.info("  Extended PAN ID: %s", self.extended_pan_id)
        LOGGER.info("  Device IEEE: %s", self.ieee)
        LOGGER.info("  Device NWK: 0x%04X", self.nwk)
        LOGGER.debug(
            "  Network key: %s",
            ":".join(
                f"{c:02x}" for c in self.state.network_information.network_key.key
            ),
        )

        if self.state.network_information.network_key.key == const.Z2M_NETWORK_KEY:
            LOGGER.warning(
                "Your network is using the insecure Zigbee2MQTT network key!"
            )

        self._watchdog_task = asyncio.create_task(self._watchdog_loop())

    async def set_tx_power(self, dbm: int) -> None:
        """
        Sets the radio TX power.
        """

        rsp = await self._znp.request(c.SYS.SetTxPower.Req(TXPower=dbm))

        if self._znp.version >= 3.30 and rsp.StatusOrPower != t.Status.SUCCESS:
            # Z-Stack 3's response indicates success or failure
            raise InvalidCommandResponse(
                f"Failed to set TX power: {t.Status(rsp.StatusOrPower)!r}", rsp
            )
        elif self._znp.version < 3.30 and rsp.StatusOrPower != dbm:
            # Old Z-Stack releases used the response status field to indicate the power
            # setting that was actually applied
            LOGGER.warning(
                "Requested TX power %d was adjusted to %d", dbm, rsp.StatusOrPower
            )

    async def form_network(self):
        """
        Clears the current config and forms a new network with a random network key,
        PAN ID, and extended PAN ID.
        """

        # First, make the settings consistent and randomly generate missing values
        channel = self.config[conf.CONF_NWK][conf.CONF_NWK_CHANNEL]
        channels = self.config[conf.CONF_NWK][conf.CONF_NWK_CHANNELS]
        pan_id = self.config[conf.CONF_NWK][conf.CONF_NWK_PAN_ID]
        extended_pan_id = self.config[conf.CONF_NWK][conf.CONF_NWK_EXTENDED_PAN_ID]
        network_key = self.config[conf.CONF_NWK][conf.CONF_NWK_KEY]

        if pan_id is None:
            pan_id = random.SystemRandom().randint(0x0001, 0xFFFE + 1)

        if extended_pan_id is None:
            extended_pan_id = ExtendedPanId(os.urandom(8))

        if network_key is None:
            network_key = t.KeyData(os.urandom(16))

        # Override `channels` with a single channel if one is explicitly set
        if channel is not None:
            channels = t.Channels.from_channel_list([channel])

        network_info = zigpy.state.NetworkInformation(
            extended_pan_id=extended_pan_id,
            pan_id=pan_id,
            nwk_update_id=self.config[conf.CONF_NWK][conf.CONF_NWK_UPDATE_ID],
            nwk_manager_id=0x0000,
            channel=channel,
            channel_mask=channels,
            security_level=5,
            network_key=zigpy.state.Key(
                key=network_key,
                tx_counter=0,
                rx_counter=0,
                seq=0,
                partner_ieee=None,
            ),
            tc_link_key=None,
            children=[],
            key_table=[],
            nwk_addresses={},
            stack_specific={"zstack": {"tclk_seed": os.urandom(16).hex()}},
        )

        node_info = zigpy.state.NodeInfo(
            nwk=0x0000,
            ieee=None,
            logical_type=zdo_t.LogicalType.Coordinator,
        )

        await self.write_network_info(network_info=network_info, node_info=node_info)
        await self._write_stack_settings(reset_if_changed=False)
        await self._znp.reset()

    def get_dst_address(self, cluster: zigpy.zcl.Cluster) -> MultiAddress:
        """
        Helper to get a dst address for bind/unbind operations.

        Allows radios to provide correct information, especially for radios which listen
        on specific endpoints only.
        """

        dst_addr = MultiAddress()
        dst_addr.addrmode = 0x03
        dst_addr.ieee = self.ieee
        dst_addr.endpoint = self._find_endpoint(
            dst_ep=cluster.endpoint.endpoint_id,
            profile=cluster.endpoint.profile_id,
            cluster=cluster.cluster_id,
        )

        return dst_addr

    @zigpy.util.retryable_request
    async def request(
        self,
        device,
        profile,
        cluster,
        src_ep,
        dst_ep,
        sequence,
        data,
        expect_reply=True,
        use_ieee=False,
    ) -> tuple[t.Status, str]:
        tx_options = c.af.TransmitOptions.SUPPRESS_ROUTE_DISC_NETWORK

        if expect_reply:
            tx_options |= c.af.TransmitOptions.ACK_REQUEST

        if use_ieee:
            destination = t.AddrModeAddress(mode=t.AddrMode.IEEE, address=device.ieee)
        else:
            destination = t.AddrModeAddress(mode=t.AddrMode.NWK, address=device.nwk)

        return await self._send_request(
            dst_addr=destination,
            dst_ep=dst_ep,
            src_ep=src_ep,
            profile=profile,
            cluster=cluster,
            sequence=sequence,
            options=tx_options,
            radius=30,
            data=data,
        )

    async def broadcast(
        self,
        profile,
        cluster,
        src_ep,
        dst_ep,
        grpid,
        radius,
        sequence,
        data,
        broadcast_address=zigpy.types.BroadcastAddress.RX_ON_WHEN_IDLE,
    ) -> tuple[t.Status, str]:
        assert grpid == 0

        return await self._send_request(
            dst_addr=t.AddrModeAddress(
                mode=t.AddrMode.Broadcast, address=broadcast_address
            ),
            dst_ep=dst_ep,
            src_ep=src_ep,
            profile=profile,
            cluster=cluster,
            sequence=sequence,
            options=c.af.TransmitOptions.NONE,
            radius=radius,
            data=data,
        )

    async def mrequest(
        self,
        group_id,
        profile,
        cluster,
        src_ep,
        sequence,
        data,
        *,
        hops=0,
        non_member_radius=3,
    ) -> tuple[t.Status, str]:
        return await self._send_request(
            dst_addr=t.AddrModeAddress(mode=t.AddrMode.Group, address=group_id),
            dst_ep=src_ep,
            src_ep=src_ep,
            profile=profile,
            cluster=cluster,
            sequence=sequence,
            options=c.af.TransmitOptions.NONE,
            radius=hops,
            data=data,
        )

    async def force_remove(self, device: zigpy.device.Device) -> None:
        """
        Attempts to forcibly remove a device from the network.
        """

        LOGGER.warning("Z-Stack does not support force remove")

    async def permit(self, time_s=60, node=None):
        """
        Permit joining the network via a specific node or via all router nodes.
        """

        LOGGER.info("Permitting joins for %d seconds", time_s)

        # If joins were permitted through a specific router, older Z-Stack builds
        # did not allow the key to be distributed unless the coordinator itself was
        # also permitting joins. This also needs to happen if we're permitting joins
        # through the coordinator itself.
        #
        # Fixed in https://github.com/Koenkk/Z-Stack-firmware/commit/efac5ee46b9b437
        if time_s == 0 or self._zstack_build_id < 20210708 or node == self.ieee:
            response = await self._znp.request_callback_rsp(
                request=c.ZDO.MgmtPermitJoinReq.Req(
                    AddrMode=t.AddrMode.NWK,
                    Dst=0x0000,
                    Duration=time_s,
                    TCSignificance=1,
                ),
                RspStatus=t.Status.SUCCESS,
                callback=c.ZDO.MgmtPermitJoinRsp.Callback(Src=0x0000, partial=True),
            )

            if response.Status != t.Status.SUCCESS:
                raise RuntimeError(f"Failed to permit joins on coordinator: {response}")

        await super().permit(time_s=time_s, node=node)

    async def permit_ncp(self, time_s: int) -> None:
        """
        Permits joins only on the coordinator.
        """

        # Z-Stack does not need any special code to do this
        pass

    async def permit_with_key(self, node: t.EUI64, code: bytes, time_s=60):
        """
        Permits a new device to join with the given IEEE and Install Code.
        """

        key = zigpy.util.convert_install_code(code)
        install_code_format = c.app_config.InstallCodeFormat.KeyDerivedFromInstallCode

        if key is None:
            raise ValueError(f"Invalid install code: {code!r}")

        await self._znp.request(
            c.AppConfig.BDBAddInstallCode.Req(
                InstallCodeFormat=install_code_format,
                IEEE=node,
                InstallCode=t.Bytes(key),
            ),
            RspStatus=t.Status.SUCCESS,
        )

        # Temporarily only allow joins that use an install code
        await self._znp.request(
            c.AppConfig.BDBSetJoinUsesInstallCodeKey.Req(
                BdbJoinUsesInstallCodeKey=True
            ),
            RspStatus=t.Status.SUCCESS,
        )

        try:
            await self.permit(time_s)
            await asyncio.sleep(time_s)
        finally:
            # Revert back to normal. The BDB config is not persistent so if this request
            # fails, we will be back to normal the next time Z-Stack resets.
            await self._znp.request(
                c.AppConfig.BDBSetJoinUsesInstallCodeKey.Req(
                    BdbJoinUsesInstallCodeKey=False
                ),
                RspStatus=t.Status.SUCCESS,
            )

    def connection_lost(self, exc):
        """
        Propagated up from UART through ZNP when the connection is lost.
        Spawns the auto-reconnect task.
        """

        LOGGER.debug("Connection lost: %s", exc)

        self.close()

        LOGGER.debug("Restarting background reconnection task")
        self._reconnect_task = asyncio.create_task(self._reconnect())

    #####################################################
    # Z-Stack message callbacks attached during startup #
    #####################################################

    def _bind_callbacks(self) -> None:
        """
        Binds all of the necessary message callbacks to their associated methods.

        Z-Stack intercepts most (but not all!) ZDO requests/responses and replaces them
        ZNP requests/responses.
        """

        self._znp.callback_for_response(
            c.AF.IncomingMsg.Callback(partial=True), self.on_af_message
        )

        # ZDO requests need to be handled explicitly, one by one
        self._znp.callback_for_response(
            c.ZDO.EndDeviceAnnceInd.Callback(partial=True),
            self.on_zdo_device_announce,
        )

        self._znp.callback_for_response(
            c.ZDO.TCDevInd.Callback.Callback(partial=True),
            self.on_zdo_tc_device_join,
        )

        self._znp.callback_for_response(
            c.ZDO.LeaveInd.Callback(partial=True), self.on_zdo_device_leave
        )

        self._znp.callback_for_response(
            c.ZDO.SrcRtgInd.Callback(partial=True), self.on_zdo_relays_message
        )

        self._znp.callback_for_response(
            c.ZDO.PermitJoinInd.Callback(partial=True), self.on_zdo_permit_join_message
        )

        # No-op handle commands that just create unnecessary WARNING logs
        self._znp.callback_for_response(
            c.ZDO.ParentAnnceRsp.Callback(partial=True),
            self.on_intentionally_unhandled_message,
        )

        self._znp.callback_for_response(
            c.ZDO.ConcentratorInd.Callback(partial=True),
            self.on_intentionally_unhandled_message,
        )

        self._znp.callback_for_response(
            c.ZDO.MgmtNWKUpdateNotify.Callback(partial=True),
            self.on_intentionally_unhandled_message,
        )

    def on_intentionally_unhandled_message(self, msg: t.CommandBase) -> None:
        """
        Some commands are unhandled but frequently sent by devices on the network. To
        reduce unnecessary logging messages, they are given an explicit callback.
        """

        pass

    def on_zdo_permit_join_message(self, msg: c.ZDO.PermitJoinInd.Callback) -> None:
        """
        Coordinator join status change message. Only sent with Z-Stack 1.2 and 3.0.
        """

        if msg.Duration == 0:
            LOGGER.info("Coordinator is not permitting joins anymore")
        else:
            LOGGER.info("Coordinator is permitting joins for %d seconds", msg.Duration)

    async def on_zdo_relays_message(self, msg: c.ZDO.SrcRtgInd.Callback) -> None:
        """
        ZDO source routing message callback
        """

        device = await self._get_or_discover_device(nwk=msg.DstAddr)

        if device is None:
            LOGGER.warning(
                "Received a ZDO message from an unknown device: %s", msg.DstAddr
            )
            return

        # `relays` is a property with a setter that emits an event
        device.relays = msg.Relays

    def on_zdo_device_announce(self, msg: c.ZDO.EndDeviceAnnceInd.Callback) -> None:
        """
        ZDO end device announcement callback
        """

        LOGGER.info("ZDO device announce: %s", msg)

        # Cancel an existing join timer so we don't double announce
        if msg.IEEE in self._join_announce_tasks:
            self._join_announce_tasks.pop(msg.IEEE).cancel()

        # Sometimes devices change their NWK when announcing so re-join it.
        self.handle_join(nwk=msg.NWK, ieee=msg.IEEE, parent_nwk=None)

        device = self.get_device(ieee=msg.IEEE)

        # We turn this back into a ZDO message and let zigpy handle it
        self._receive_zdo_message(
            cluster=ZDOCmd.Device_annce,
            tsn=0xFF,
            sender=device,
            NWKAddr=msg.NWK,
            IEEEAddr=msg.IEEE,
            Capability=msg.Capabilities,
        )

    def on_zdo_tc_device_join(self, msg: c.ZDO.TCDevInd.Callback) -> None:
        """
        ZDO trust center device join callback.
        """

        LOGGER.info("TC device join: %s", msg)

        # Perform route discovery (just in case) when a device joins the network so that
        # we can begin initialization as soon as possible.
        asyncio.create_task(self._discover_route(msg.SrcNwk))

        if msg.SrcIEEE in self._join_announce_tasks:
            self._join_announce_tasks.pop(msg.SrcIEEE).cancel()

        # If the device already exists, immediately trigger a join to update its NWK.
        try:
            self.get_device(ieee=msg.SrcIEEE)
        except KeyError:
            pass
        else:
            self.handle_join(
                nwk=msg.SrcNwk,
                ieee=msg.SrcIEEE,
                parent_nwk=msg.ParentNwk,
            )
            return

        # Some devices really don't like zigpy beginning its initialization process
        # before the device has announced itself. Wait a second or two before calling
        # `handle_join`, just in case the device announces itself first.
        self._join_announce_tasks[msg.SrcIEEE] = asyncio.get_running_loop().call_later(
            DEVICE_JOIN_MAX_DELAY,
            lambda: self.handle_join(
                nwk=msg.SrcNwk,
                ieee=msg.SrcIEEE,
                parent_nwk=msg.ParentNwk,
            ),
        )

    def on_zdo_device_leave(self, msg: c.ZDO.LeaveInd.Callback) -> None:
        LOGGER.info("ZDO device left: %s", msg)
        self.handle_leave(nwk=msg.NWK, ieee=msg.IEEE)

    async def on_af_message(self, msg: c.AF.IncomingMsg.Callback) -> None:
        """
        Handler for all non-ZDO messages.
        """

        device = await self._get_or_discover_device(nwk=msg.SrcAddr)

        if device is None:
            LOGGER.warning(
                "Received an AF message from an unknown device: %s", msg.SrcAddr
            )
            return

        device.radio_details(lqi=msg.LQI, rssi=None)

        # XXX: Is it possible to receive messages on non-assigned endpoints?
        if msg.DstEndpoint in self.zigpy_device.endpoints:
            profile = self.zigpy_device.endpoints[msg.DstEndpoint].profile_id
        else:
            LOGGER.warning("Received a message on an unregistered endpoint: %s", msg)
            profile = zigpy.profiles.zha.PROFILE_ID

        self.handle_message(
            sender=device,
            profile=profile,
            cluster=msg.ClusterId,
            src_ep=msg.SrcEndpoint,
            dst_ep=msg.DstEndpoint,
            message=msg.Data,
        )

    ####################
    # Internal methods #
    ####################

    @property
    def _zstack_build_id(self) -> t.uint32_t:
        """
        Z-Stack build ID, more recently the build date.
        """

        return self._version_rsp.CodeRevision

    @property
    def zigpy_device(self) -> zigpy.device.Device:
        """
        Reference to zigpy device 0x0000, the coordinator.
        """

        return self.devices[self.ieee]

    @property
    def znp_config(self) -> conf.ConfigType:
        """
        Shortcut property to access the ZNP radio config.
        """

        return self.config[conf.CONF_ZNP_CONFIG]

    async def _watchdog_loop(self):
        """
        Watchdog loop to periodically test if Z-Stack is still running.
        """

        LOGGER.debug("Starting watchdog loop")

        while True:
            await asyncio.sleep(WATCHDOG_PERIOD)

            # No point in trying to test the port if it's already disconnected
            if self._znp is None:
                break

            try:
                await self._znp.request(c.SYS.Ping.Req())
            except Exception as e:
                LOGGER.error(
                    "Watchdog check failed",
                    exc_info=e,
                )

                # Treat the watchdog failure as a disconnect
                self.connection_lost(e)

                return

    @combine_concurrent_calls
    async def _get_or_discover_device(self, nwk: t.NWK) -> zigpy.device.Device | None:
        """
        Finds a device by its NWK address. If a device does not exist in the zigpy
        database, attempt to look up its new NWK address. If it does not exist in the
        zigpy database, treat the device as a new join.
        """

        try:
            return self.get_device(nwk=nwk)
        except KeyError:
            pass

        LOGGER.debug("Device with NWK 0x%04X not in database", nwk)

        try:
            # XXX: Multiple responses may arrive but we only use the first one
            ieee_addr_rsp = await self._znp.request_callback_rsp(
                request=c.ZDO.IEEEAddrReq.Req(
                    NWK=nwk,
                    RequestType=c.zdo.AddrRequestType.SINGLE,
                    StartIndex=0,
                ),
                RspStatus=t.Status.SUCCESS,
                callback=c.ZDO.IEEEAddrRsp.Callback(
                    partial=True,
                    NWK=nwk,
                ),
                timeout=5,  # We don't want to wait forever
            )
        except asyncio.TimeoutError:
            return None

        ieee = ieee_addr_rsp.IEEE

        try:
            device = self.get_device(ieee=ieee)
        except KeyError:
            LOGGER.debug("Treating unknown device as a new join")
            self.handle_join(nwk=nwk, ieee=ieee, parent_nwk=None)

            return self.get_device(ieee=ieee)

        # The `Device` object could have been updated while this coroutine is running
        if device.nwk == nwk:
            return device

        LOGGER.warning(
            "Device %s changed its NWK from %s to %s",
            device.ieee,
            device.nwk,
            nwk,
        )

        # Notify zigpy of the change
        self.handle_join(nwk=nwk, ieee=ieee, parent_nwk=None)

        # `handle_join` will update the NWK
        assert device.nwk == nwk

        return device

    async def _set_led_mode(self, *, led: t.uint8_t, mode: c.util.LEDMode) -> None:
        """
        Attempts to set the provided LED's mode. A Z-Stack bug causes the underlying
        command to never receive a response if the board has no LEDs, requiring this
        wrapper function to prevent the command from taking many seconds to time out.
        """

        # XXX: If Z-Stack is not compiled with HAL_LED, it will just not respond at all
        try:
            async with async_timeout.timeout(0.5):
                await self._znp.request(
                    c.UTIL.LEDControl.Req(LED=led, Mode=mode),
                    RspStatus=t.Status.SUCCESS,
                )
        except (asyncio.TimeoutError, CommandNotRecognized):
            LOGGER.info("This build of Z-Stack does not appear to support LED control")

    async def _migrate_nvram(self):
        """
        Migrates NVRAM entries using the `ZIGPY_ZNP_MIGRATION_ID` NVRAM item.
        """

        try:
            migration_id = await self._znp.nvram.osal_read(
                OsalNvIds.ZIGPY_ZNP_MIGRATION_ID, item_type=t.uint8_t
            )
        except KeyError:
            migration_id = 0

        initial_migration_id = migration_id

        # Migration 1: empty `ADDRMGR` entries are version-dependent and were improperly
        #              written for CC253x devices.
        #
        #              This migration is stateless and can safely be run more than once:
        #              the only downside is that startup times increase by 10s on newer
        #              coordinators, which is why the migration ID is persisted.
        if migration_id < 1:
            try:
                entries = await security.read_addr_manager_entries(self._znp)
            except KeyError:
                pass
            else:
                fixed_entries = []

                for entry in entries:
                    if entry.extAddr != t.EUI64.convert("FF:FF:FF:FF:FF:FF:FF:FF"):
                        fixed_entries.append(entry)
                    elif self._znp.version == 3.30:
                        fixed_entries.append(const.EMPTY_ADDR_MGR_ENTRY_ZSTACK3)
                    else:
                        fixed_entries.append(const.EMPTY_ADDR_MGR_ENTRY_ZSTACK1)

                if entries != fixed_entries:
                    LOGGER.warning(
                        "Repairing %d invalid empty address manager entries (total %d)",
                        sum(i != j for i, j in zip(entries, fixed_entries)),
                        len(entries),
                    )
                    await security.write_addr_manager_entries(self._znp, fixed_entries)

            migration_id = 1

        if initial_migration_id != migration_id:
            await self._znp.nvram.osal_write(
                OsalNvIds.ZIGPY_ZNP_MIGRATION_ID, t.uint8_t(migration_id), create=True
            )
            await self._znp.reset()

    async def _write_stack_settings(self, *, reset_if_changed: bool) -> None:
        """
        Writes network-independent Z-Stack settings to NVRAM.
        If no settings actually change, no reset will be performed.
        """

        # It's better to be explicit than rely on the NVRAM defaults
        settings = {
            OsalNvIds.LOGICAL_TYPE: t.DeviceLogicalType.Coordinator,
            # Source routing
            OsalNvIds.CONCENTRATOR_ENABLE: t.Bool(True),
            OsalNvIds.CONCENTRATOR_DISCOVERY: t.uint8_t(120),
            OsalNvIds.CONCENTRATOR_RC: t.Bool(True),
            OsalNvIds.SRC_RTG_EXPIRY_TIME: t.uint8_t(255),
            OsalNvIds.NWK_CHILD_AGE_ENABLE: t.Bool(False),
            # Default is 20 in Z-Stack 3.0.1, 30 in Z-Stack 3/4
            OsalNvIds.BCAST_DELIVERY_TIME: t.uint8_t(30),
            # We want to receive all ZDO callbacks to proxy them back to zigpy
            OsalNvIds.ZDO_DIRECT_CB: t.Bool(True),
        }

        any_changed = False

        for nvid, value in settings.items():
            try:
                current_value = await self._znp.nvram.osal_read(
                    nvid, item_type=type(value)
                )
            except InvalidCommandResponse:
                current_value = None

            # There is no point in issuing a write if the value will not change
            if current_value != value:
                await self._znp.nvram.osal_write(nvid, value)
                any_changed = True

        if reset_if_changed and any_changed:
            # Reset to make the above NVRAM writes take effect
            await self._znp.reset()

    @contextlib.asynccontextmanager
    async def _limit_concurrency(self):
        """
        Async context manager that prevents devices from being overwhelmed by requests.
        Mainly a thin wrapper around `asyncio.Semaphore` that logs when it has to wait.
        """

        # Allow sending some requests before the application has fully started
        if self._concurrent_requests_semaphore is None:
            yield
            return

        start_time = time.time()
        was_locked = self._concurrent_requests_semaphore.locked()

        if was_locked:
            self._currently_waiting_requests += 1
            LOGGER.debug(
                "Max concurrency reached, delaying requests (%s enqueued)",
                self._currently_waiting_requests,
            )

        try:
            async with self._concurrent_requests_semaphore:
                if was_locked:
                    LOGGER.debug(
                        "Previously delayed request is now running, "
                        "delayed by %0.2f seconds",
                        time.time() - start_time,
                    )

                yield
        finally:
            if was_locked:
                self._currently_waiting_requests -= 1

    def _receive_zdo_message(
        self,
        cluster: ZDOCmd,
        *,
        tsn: t.uint8_t,
        sender: zigpy.device.Device,
        **zdo_kwargs,
    ) -> None:
        """
        Internal method that is mainly called by our ZDO request/response converters to
        receive a "fake" ZDO message constructed from a cluster and args/kwargs.
        """

        field_names, field_types = ZDO_CLUSTERS[cluster]
        assert set(zdo_kwargs) == set(field_names)

        # Type cast all of the field args and kwargs
        zdo_args = [t(zdo_kwargs[name]) for name, t in zip(field_names, field_types)]
        message = t.serialize_list([t.uint8_t(tsn)] + zdo_args)

        LOGGER.debug("Pretending we received a ZDO message: %s", message)

        self.handle_message(
            sender=sender,
            profile=zigpy.profiles.zha.PROFILE_ID,
            cluster=cluster,
            src_ep=ZDO_ENDPOINT,
            dst_ep=ZDO_ENDPOINT,
            message=message,
        )

    async def _reconnect(self) -> None:
        """
        Endlessly tries to reconnect to the currently configured radio.

        Relies on the fact that `self.startup()` only modifies `self` upon a successful
        connection to be essentially stateless.
        """

        for attempt in itertools.count(start=1):
            LOGGER.debug(
                "Trying to reconnect to %s, attempt %d",
                self._config[conf.CONF_DEVICE][conf.CONF_DEVICE_PATH],
                attempt,
            )

            try:
                await self._startup()
                return
            except asyncio.CancelledError:
                raise
            except Exception as e:
                LOGGER.error("Failed to reconnect", exc_info=e)

                if self._znp is not None:
                    self._znp.close()
                    self._znp = None

                await asyncio.sleep(
                    self._config[conf.CONF_ZNP_CONFIG][
                        conf.CONF_AUTO_RECONNECT_RETRY_DELAY
                    ]
                )

    async def _register_endpoints(self) -> None:
        """
        Registers the Zigbee endpoints required to communicate with various devices.
        """

        await self._znp.request(
            c.AF.Register.Req(
                Endpoint=ZHA_ENDPOINT,
                ProfileId=zigpy.profiles.zha.PROFILE_ID,
                DeviceId=zigpy.profiles.zha.DeviceType.IAS_CONTROL,
                DeviceVersion=0b0000,
                LatencyReq=c.af.LatencyReq.NoLatencyReqs,
                InputClusters=[
                    clusters.general.Basic.cluster_id,
                    clusters.general.Ota.cluster_id,
                ],
                OutputClusters=[
                    clusters.security.IasZone.cluster_id,
                    clusters.security.IasWd.cluster_id,
                ],
            ),
            RspStatus=t.Status.SUCCESS,
        )

        await self._znp.request(
            c.AF.Register.Req(
                Endpoint=ZLL_ENDPOINT,
                ProfileId=zigpy.profiles.zll.PROFILE_ID,
                DeviceId=zigpy.profiles.zll.DeviceType.CONTROLLER,
                DeviceVersion=0b0000,
                LatencyReq=c.af.LatencyReq.NoLatencyReqs,
                InputClusters=[clusters.general.Basic.cluster_id],
                OutputClusters=[],
            ),
            RspStatus=t.Status.SUCCESS,
        )

    async def load_network_info(self, *, load_devices=False) -> None:
        """
        Loads network information from NVRAM.
        """

        await self._znp.load_network_info(load_devices=load_devices)

        self.state.node_information = self._znp.node_info
        self.state.network_information = self._znp.network_info

    async def write_network_info(
        self,
        *,
        network_info: zigpy.state.NetworkInformation,
        node_info: zigpy.state.NodeInfo,
    ) -> None:
        """
        Writes network and node state to NVRAM.
        """

        return await self._znp.write_network_info(
            network_info=network_info, node_info=node_info
        )

    def _find_endpoint(self, dst_ep: int, profile: int, cluster: int) -> int:
        """
        Zigpy defaults to sending messages with src_ep == dst_ep. This does not work
        with all versions of Z-Stack, which requires endpoints to be registered
        explicitly on startup.
        """

        if dst_ep == ZDO_ENDPOINT:
            return ZDO_ENDPOINT

        # Newer Z-Stack releases ignore profiles and will work properly with endpoint 1
        if self._zstack_build_id >= 20210708:
            return ZHA_ENDPOINT

        # Always fall back to endpoint 1
        candidates = [ZHA_ENDPOINT]

        for ep_id, endpoint in self.zigpy_device.endpoints.items():
            if ep_id == ZDO_ENDPOINT:
                continue

            if endpoint.profile_id != profile:
                continue

            # An exact match, no need to continue further
            # TODO: pass in `is_server_cluster` or something similar
            if cluster in endpoint.out_clusters or cluster in endpoint.in_clusters:
                return endpoint.endpoint_id

            # Otherwise, keep track of the candidate cluster
            # if we don't find anything better
            candidates.append(endpoint.endpoint_id)

        return candidates[-1]

    async def _send_zdo_request(
        self, dst_addr, dst_ep, src_ep, cluster, sequence, options, radius, data
    ):
        """
        Zigpy doesn't send ZDO requests via TI's ZDO_* MT commands, so it will never
        receive a reply because ZNP intercepts ZDO replies, never sends a DataConfirm,
        and instead replies with one of its ZDO_* MT responses.

        This method translates the ZDO_* MT response into one zigpy can handle.
        """

        LOGGER.debug(
            "Intercepted a ZDO request: dst_addr=%s, dst_ep=%s, src_ep=%s, "
            "cluster=%s, sequence=%s, options=%s, radius=%s, data=%s",
            dst_addr,
            dst_ep,
            src_ep,
            cluster,
            sequence,
            options,
            radius,
            data,
        )

        assert dst_ep == ZDO_ENDPOINT

        # Deserialize the ZDO request
        zdo_hdr, data = ZDOHeader.deserialize(cluster, data)
        field_names, field_types = ZDO_CLUSTERS[cluster]
        zdo_args, _ = list_deserialize(data, field_types)
        zdo_kwargs = dict(zip(field_names, zdo_args))

        # TODO: Check out `ZDO.MsgCallbackRegister`

        if cluster not in ZDO_CONVERTERS:
            LOGGER.error(
                "ZDO converter for cluster %s has not been implemented!"
                " Please open a GitHub issue and attach a debug log:"
                " https://github.com/zigpy/zigpy-znp/issues/new",
                cluster,
            )
            raise RuntimeError("No ZDO converter")

        # Call the converter with the ZDO request's kwargs
        req_factory, rsp_factory, zdo_rsp_factory = ZDO_CONVERTERS[cluster]
        request = req_factory(dst_addr, **zdo_kwargs)
        callback = rsp_factory(dst_addr)

        LOGGER.debug(
            "Intercepted AP ZDO request %s(%s) and replaced with %s",
            cluster,
            zdo_kwargs,
            request,
        )

        # The coordinator responds to broadcasts
        if dst_addr.mode == t.AddrMode.Broadcast:
            callback = callback.replace(Src=0x0000)

        async with async_timeout.timeout(ZDO_REQUEST_TIMEOUT):
            response = await self._znp.request_callback_rsp(
                request=request, RspStatus=t.Status.SUCCESS, callback=callback
            )

        # We should only send zigpy unicast responses
        if dst_addr.mode == t.AddrMode.NWK:
            zdo_rsp_cluster, zdo_response_kwargs = zdo_rsp_factory(response)

            self._receive_zdo_message(
                cluster=zdo_rsp_cluster,
                tsn=sequence,
                sender=self.get_device(nwk=dst_addr.address),
                **zdo_response_kwargs,
            )

        return response

    async def _send_request_raw(
        self,
        dst_addr,
        dst_ep,
        src_ep,
        profile,
        cluster,
        sequence,
        options,
        radius,
        data,
        *,
        relays=None,
    ):
        """
        Used by `request`/`mrequest`/`broadcast` to send a request.
        Picks the correct request sending mechanism and fixes endpoint information.
        """

        # ZDO requests must be handled by the translation layer, since Z-Stack will
        # "steal" the responses
        if dst_ep == ZDO_ENDPOINT:
            return await self._send_zdo_request(
                dst_addr, dst_ep, src_ep, cluster, sequence, options, radius, data
            )

        # Zigpy just sets src == dst, which doesn't work for devices with many endpoints
        # We pick ours based on the registered endpoints when using an older firmware
        src_ep = self._find_endpoint(dst_ep=dst_ep, profile=profile, cluster=cluster)

        if relays is None:
            request = c.AF.DataRequestExt.Req(
                DstAddrModeAddress=dst_addr,
                DstEndpoint=dst_ep,
                DstPanId=0x0000,
                SrcEndpoint=src_ep,
                ClusterId=cluster,
                TSN=sequence,
                Options=options,
                Radius=radius,
                Data=data,
            )
        else:
            request = c.AF.DataRequestSrcRtg.Req(
                DstAddr=dst_addr.address,
                DstEndpoint=dst_ep,
                SrcEndpoint=src_ep,
                ClusterId=cluster,
                TSN=sequence,
                Options=options,
                Radius=radius,
                SourceRoute=relays,  # force the packet to go through specific parents
                Data=data,
            )

        if dst_addr.mode == t.AddrMode.Broadcast:
            # Broadcasts will not receive a confirmation
            response = await self._znp.request(
                request=request, RspStatus=t.Status.SUCCESS
            )
        else:
            async with async_timeout.timeout(DATA_CONFIRM_TIMEOUT):
                # Shield from cancellation to prevent requests that time out in higher
                # layers from missing expected responses
                response = await asyncio.shield(
                    self._znp.request_callback_rsp(
                        request=request,
                        RspStatus=t.Status.SUCCESS,
                        callback=c.AF.DataConfirm.Callback(
                            partial=True,
                            TSN=sequence,
                            # XXX: can this ever not match?
                            # Endpoint=src_ep,
                        ),
                        # Multicasts eventually receive a confirmation but waiting for
                        # it is unnecessary
                        background=(dst_addr.mode == t.AddrMode.Group),
                    )
                )

                # Both the callback and the response can have an error status
                if response.Status != t.Status.SUCCESS:
                    raise InvalidCommandResponse(
                        f"Unsuccessful request status code: {response.Status!r}",
                        response,
                    )

        return response

    @combine_concurrent_calls
    async def _discover_route(self, nwk: t.NWK) -> None:
        """
        Instructs the coordinator to re-discover routes to the provided NWK.
        Runs concurrently and at most once per NWK, even if called multiple times.
        """

        # Route discovery with Z-Stack 1.2 and Z-Stack 3.0.2 on the CC2531 doesn't
        # appear to work very well (Z2M#2901)
        if self._znp.version < 3.30:
            return

        await self._znp.request(
            c.ZDO.ExtRouteDisc.Req(
                Dst=nwk,
                Options=c.zdo.RouteDiscoveryOptions.UNICAST,
                Radius=30,
            ),
        )

        await asyncio.sleep(0.1 * 13)

    async def _send_request(
        self,
        dst_addr,
        dst_ep,
        src_ep,
        profile,
        cluster,
        sequence,
        options,
        radius,
        data,
    ) -> tuple[t.Status, str]:
        """
        Fault-tolerant wrapper around `_send_request_raw` that transparently attempts to
        repair routes and contact the device through other methods when Z-Stack errors
        are encountered.
        """

        try:
            if dst_addr.mode == t.AddrMode.NWK:
                device = self.get_device(nwk=dst_addr.address)
            elif dst_addr.mode == t.AddrMode.IEEE:
                device = self.get_device(ieee=dst_addr.address)
            else:
                device = None
        except KeyError:
            # Sometimes a request is sent to a device not in the database. This should
            # work, the device object is only for recovery.
            device = None

        status = None
        response = None
        association = None
        force_relays = None

        tried_assoc_remove = False
        tried_route_discovery = False
        tried_last_good_route = False
        tried_ieee_address = False

        # Don't release the concurrency-limiting semaphore until we are done trying.
        # There is no point in allowing requests to take turns getting buffer errors.
        try:
            async with self._limit_concurrency():
                for attempt in range(REQUEST_MAX_RETRIES):
                    try:
                        # ZDO requests do not generate `AF.DataConfirm` messages
                        # indicating that a route is missing so we need to explicitly
                        # check for one.
                        if (
                            dst_ep == ZDO_ENDPOINT
                            and dst_addr.mode == t.AddrMode.NWK
                            and dst_addr.address != 0x0000
                        ):
                            route_status = await self._znp.request(
                                c.ZDO.ExtRouteChk.Req(
                                    Dst=dst_addr.address,
                                    RtStatus=c.zdo.RouteStatus.ACTIVE,
                                    Options=(
                                        c.zdo.RouteOptions.MTO_ROUTE
                                        | c.zdo.RouteOptions.NO_ROUTE_CACHE
                                    ),
                                )
                            )

                            if route_status.Status != c.zdo.RoutingStatus.SUCCESS:
                                await self._discover_route(dst_addr.address)

                        response = await self._send_request_raw(
                            dst_addr=dst_addr,
                            dst_ep=dst_ep,
                            src_ep=src_ep,
                            profile=profile,
                            cluster=cluster,
                            sequence=sequence,
                            options=options,
                            radius=radius,
                            data=data,
                            relays=force_relays,
                        )
                        break
                    except InvalidCommandResponse as e:
                        status = e.response.Status

                        if status not in REQUEST_RETRYABLE_ERRORS:
                            raise

                        # Just retry if:
                        #  - this is our first failure
                        #  - the error is transient and is not a routing issue
                        #  - or we are not sending a unicast request
                        if (
                            attempt == 0
                            or status in REQUEST_TRANSIENT_ERRORS
                            or dst_addr.mode not in (t.AddrMode.NWK, t.AddrMode.IEEE)
                        ):
                            LOGGER.debug(
                                "Request failed (%s), retry attempt %s of %s",
                                e,
                                attempt + 1,
                                REQUEST_MAX_RETRIES,
                            )
                            await asyncio.sleep(3 * REQUEST_ERROR_RETRY_DELAY)
                            continue

                        # If we can't contact the device by forcing a specific route,
                        # there is not point in trying this more than once.
                        if tried_last_good_route and force_relays is not None:
                            force_relays = None

                        # If we fail to contact the device with its IEEE address, don't
                        # try again.
                        if (
                            tried_ieee_address
                            and dst_addr.mode == t.AddrMode.IEEE
                            and device is not None
                        ):
                            dst_addr = t.AddrModeAddress(
                                mode=t.AddrMode.NWK,
                                address=device.nwk,
                            )

                        # Child aging is disabled so if a child switches parents from
                        # the coordinator to another router, we will not be able to
                        # re-discover a route to it. We have to manually drop the child
                        # to do this.
                        if (
                            status == t.Status.MAC_TRANSACTION_EXPIRED
                            and device is not None
                            and association is None
                            and not tried_assoc_remove
                            and self._znp.version >= 3.30
                        ):
                            association = await self._znp.request(
                                c.UTIL.AssocGetWithAddress.Req(
                                    IEEE=device.ieee,
                                    NWK=0x0000,  # IEEE takes priority
                                )
                            )

                            if (
                                association.Device.nodeRelation
                                != c.util.NodeRelation.NOTUSED
                            ):
                                try:
                                    await self._znp.request(
                                        c.UTIL.AssocRemove.Req(IEEE=device.ieee)
                                    )
                                    tried_assoc_remove = True

                                    # Route discovery must be performed right after
                                    await self._discover_route(device.nwk)
                                except CommandNotRecognized:
                                    LOGGER.debug(
                                        "The UTIL.AssocRemove command is available only"
                                        " in Z-Stack 3 releases built after 20201017"
                                    )
                        elif not tried_last_good_route and device is not None:
                            # `ZDO.SrcRtgInd` callbacks tell us the last path taken by
                            # messages from the device back to the coordinator. Sending
                            # packets backwards via this same route may work.
                            force_relays = (device.relays or [])[::-1]
                            tried_last_good_route = True
                        elif (
                            not tried_route_discovery
                            and dst_addr.mode == t.AddrMode.NWK
                        ):
                            # If that doesn't work, try re-discovering the route.
                            # While we can in theory poll and wait until it is fixed,
                            # letting the retry mechanism deal with it simpler.
                            await self._discover_route(dst_addr.address)
                            tried_route_discovery = True
                        elif (
                            not tried_ieee_address
                            and device is not None
                            and dst_addr.mode == t.AddrMode.NWK
                        ):
                            # Try using the device's IEEE address instead of its NWK.
                            # If it works, the NWK will be updated when relays arrive.
                            tried_ieee_address = True
                            dst_addr = t.AddrModeAddress(
                                mode=t.AddrMode.IEEE,
                                address=device.ieee,
                            )

                        LOGGER.debug(
                            "Request failed (%s), retry attempt %s of %s",
                            e,
                            attempt + 1,
                            REQUEST_MAX_RETRIES,
                        )

                        # We've tried everything already so at this point just wait
                        await asyncio.sleep(REQUEST_ERROR_RETRY_DELAY)
                else:
                    raise DeliveryError(
                        f"Request failed after {REQUEST_MAX_RETRIES} attempts:"
                        f" {status!r}"
                    )
        finally:
            # We *must* re-add the device association if we previously removed it but
            # the request still failed. Otherwise, it may be a direct child and we will
            # not be able to find it again.
            if tried_assoc_remove and response is None:
                await self._znp.request(
                    c.UTIL.AssocAdd.Req(
                        NWK=association.Device.shortAddr,
                        IEEE=device.ieee,
                        NodeRelation=association.Device.nodeRelation,
                    )
                )

        if response.Status != t.Status.SUCCESS:
            return response.Status, "Failed to send request"

        return response.Status, "Sent request successfully"



from zigpy.exceptions import DeliveryError


class InvalidFrame(ValueError):
    pass


class SecurityError(Exception):
    pass


class CommandNotRecognized(Exception):
    pass


class InvalidCommandResponse(DeliveryError):
    def __init__(self, message, response):
        super().__init__(message)
        self.response = response
